import traceback
from importlib import invalidate_caches, util as imp_util
import inspect
import os
import re
import sys
import types
import uuid
from threading import Lock

import yaml

from pathlib import Path

from conan.api.output import ConanOutput
from conan.tools.cmake import cmake_layout
from conan.tools.google import bazel_layout
from conan.tools.microsoft import vs_layout
from conan.internal.errors import conanfile_exception_formatter, NotFoundException
from conan.errors import ConanException
from conan.internal.model.conan_file import ConanFile
from conan.internal.model.options import Options
from conan.api.model import RecipeReference
from conan.internal.paths import DATA_YML
from conan.internal.model.version_range import validate_conan_version
from conan.internal.util.config_parser import TextINIParse
from conan.internal.util.files import load, chdir, load_user_encoded


class ConanFileLoader:

    def __init__(self, pyreq_loader=None, conanfile_helpers=None):
        self._pyreq_loader = pyreq_loader
        self._cached_conanfile_classes = {}
        self._conanfile_helpers = conanfile_helpers
        invalidate_caches()

    def load_basic(self, conanfile_path, graph_lock=None, display="", remotes=None,
                   update=None, check_update=None):
        """ loads a conanfile basic object without evaluating anything
        """
        return self.load_basic_module(conanfile_path, graph_lock, display, remotes,
                                      update, check_update)[0]

    def load_basic_module(self, conanfile_path, graph_lock=None, display="", remotes=None,
                          update=None, check_update=None, tested_python_requires=None):
        """ loads a conanfile basic object without evaluating anything, returns the module too
        """
        cached = self._cached_conanfile_classes.get(conanfile_path)
        if cached:
            conanfile = cached[0](display)
            conanfile._conan_helpers = self._conanfile_helpers
            if hasattr(conanfile, "init") and callable(conanfile.init):
                with conanfile_exception_formatter(conanfile, "init"):
                    conanfile.init()
            return conanfile, cached[1]

        try:
            module, conanfile = _parse_conanfile(conanfile_path)
            if isinstance(tested_python_requires, RecipeReference):
                if getattr(conanfile, "python_requires", None) == "tested_reference_str":
                    conanfile.python_requires = tested_python_requires.repr_notime()
            elif tested_python_requires:
                if getattr(conanfile, "python_requires", None) != "tested_reference_str":
                    ConanOutput().warning("test_package/conanfile.py should declare 'python_requires"
                                          " = \"tested_reference_str\"'", warn_tag="deprecated")
                conanfile.python_requires = tested_python_requires

            if self._pyreq_loader:
                self._pyreq_loader.load_py_requires(conanfile, self, graph_lock, remotes,
                                                    update, check_update)

            conanfile.recipe_folder = os.path.dirname(conanfile_path)
            conanfile.recipe_path = Path(conanfile.recipe_folder)

            # Load and populate dynamic fields from the data file
            conan_data = self._load_data(conanfile_path)
            conanfile.conan_data = conan_data

            self._cached_conanfile_classes[conanfile_path] = (conanfile, module)
            result = conanfile(display)

            result._conan_helpers = self._conanfile_helpers
            if hasattr(result, "init") and callable(result.init):
                with conanfile_exception_formatter(result, "init"):
                    result.init()
            return result, module
        except ConanException as e:
            raise ConanException("Error loading conanfile at '{}': {}".format(conanfile_path, e))

    @staticmethod
    def _load_data(conanfile_path):
        data_path = os.path.join(os.path.dirname(conanfile_path), DATA_YML)
        if not os.path.exists(data_path):
            return None

        try:
            data = yaml.safe_load(load(data_path))
        except Exception as e:
            raise ConanException("Invalid yml format at {}: {}".format(DATA_YML, e))

        return data or {}

    def load_named(self, conanfile_path, name, version, user, channel, graph_lock=None,
                   remotes=None, update=None, check_update=None, tested_python_requires=None):
        """ loads the basic conanfile object and evaluates its name and version
        """
        conanfile, _ = self.load_basic_module(conanfile_path, graph_lock, remotes=remotes,
                                              update=update, check_update=check_update,
                                              tested_python_requires=tested_python_requires)

        # Export does a check on existing name & version
        if name:
            if conanfile.name and name != conanfile.name:
                raise ConanException("Package recipe with name %s!=%s" % (name, conanfile.name))
            conanfile.name = name

        if version:
            if conanfile.version and version != conanfile.version:
                raise ConanException("Package recipe with version %s!=%s"
                                     % (version, conanfile.version))
            conanfile.version = version

        if user:
            if conanfile.user and user != conanfile.user:
                raise ConanException("Package recipe with user %s!=%s"
                                     % (user, conanfile.user))
            conanfile.user = user

        if channel:
            if conanfile.channel and channel != conanfile.channel:
                raise ConanException("Package recipe with channel %s!=%s"
                                     % (channel, conanfile.channel))
            conanfile.channel = channel

        if conanfile.channel and not conanfile.user:
            raise ConanException(f"{conanfile_path}: Can't specify channel '{conanfile.channel}' without user")

        if hasattr(conanfile, "set_name"):
            with conanfile_exception_formatter("conanfile.py", "set_name"):
                conanfile.set_name()
        if hasattr(conanfile, "set_version"):
            with conanfile_exception_formatter("conanfile.py", "set_version"):
                conanfile.set_version()

        return conanfile

    def load_export(self, conanfile_path, name, version, user, channel, graph_lock=None,
                    remotes=None):
        """ loads the conanfile and evaluates its name, version, and enforce its existence
        """
        conanfile = self.load_named(conanfile_path, name, version, user, channel, graph_lock,
                                    remotes=remotes)
        if not conanfile.name:
            raise ConanException("conanfile didn't specify name")
        if not conanfile.version:
            raise ConanException("conanfile didn't specify version")

        ref = RecipeReference(conanfile.name, conanfile.version, conanfile.user, conanfile.channel)
        conanfile.display_name = str(ref)
        conanfile.output.scope = conanfile.display_name
        return conanfile

    def load_consumer(self, conanfile_path, name=None, version=None, user=None,
                      channel=None, graph_lock=None,  remotes=None, update=None, check_update=None,
                      tested_python_requires=None):
        """ loads a conanfile.py in user space. Might have name/version or not
        """
        conanfile = self.load_named(conanfile_path, name, version, user, channel, graph_lock,
                                    remotes, update, check_update,
                                    tested_python_requires=tested_python_requires)

        ref = RecipeReference(conanfile.name, conanfile.version, conanfile.user, conanfile.channel)
        if str(ref):
            conanfile.display_name = "%s (%s)" % (os.path.basename(conanfile_path), str(ref))
        else:
            conanfile.display_name = os.path.basename(conanfile_path)
        conanfile.output.scope = conanfile.display_name
        conanfile._conan_is_consumer = True
        return conanfile

    def load_conanfile(self, conanfile_path, ref, graph_lock=None, remotes=None,
                       update=None, check_update=None):
        """ load a conanfile with a full reference, name, version, user and channel are obtained
        from the reference, not evaluated. Main way to load from the cache
        """
        try:
            conanfile, _ = self.load_basic_module(conanfile_path, graph_lock, str(ref), remotes,
                                                  update=update, check_update=check_update)
        except Exception as e:
            raise ConanException("%s: Cannot load recipe.\n%s" % (str(ref), str(e)))

        conanfile.name = ref.name
        conanfile.version = str(ref.version)
        conanfile.user = ref.user
        conanfile.channel = ref.channel
        return conanfile

    def load_conanfile_txt(self, conan_txt_path):
        if not os.path.exists(conan_txt_path):
            raise NotFoundException("Conanfile not found!")

        try:
            contents = load_user_encoded(conan_txt_path)
        except Exception as e:
            raise ConanException(f"Cannot load conanfile.txt:\n{e}")
        path, basename = os.path.split(conan_txt_path)
        display_name = basename
        conanfile = self._parse_conan_txt(contents, path, display_name)
        conanfile._conan_helpers = self._conanfile_helpers
        conanfile._conan_is_consumer = True
        return conanfile

    @staticmethod
    def _parse_conan_txt(contents, path, display_name):
        conanfile = ConanFile(display_name)

        try:
            parser = ConanFileTextLoader(contents)
        except Exception as e:
            raise ConanException("%s:\n%s" % (path, str(e)))
        for reference in parser.requirements:
            conanfile.requires(reference)
        for build_reference in parser.tool_requirements:
            # TODO: Improve this interface
            conanfile.requires.tool_require(build_reference)
        for ref in parser.test_requirements:
            # TODO: Improve this interface
            conanfile.requires.test_require(ref)

        if parser.layout:
            layout_method = {"cmake_layout": cmake_layout,
                             "vs_layout": vs_layout,
                             "bazel_layout": bazel_layout}.get(parser.layout)
            if not layout_method:
                raise ConanException("Unknown predefined layout '{}' declared in "
                                     "conanfile.txt".format(parser.layout))

            def layout(_self):
                layout_method(_self)

            conanfile.layout = types.MethodType(layout, conanfile)

        conanfile.generators = parser.generators
        try:
            conanfile.options = Options.loads(parser.options)
        except Exception:
            raise ConanException("Error while parsing [options] in conanfile.txt\n"
                                 "Options should be specified as 'pkg/*:option=value'")
        return conanfile

    def load_virtual(self, requires=None, tool_requires=None, python_requires=None, graph_lock=None,
                     remotes=None, update=None, check_updates=None):
        # If user don't specify namespace in options, assume that it is
        # for the reference (keep compatibility)
        conanfile = ConanFile(display_name="cli")
        conanfile._conan_helpers = self._conanfile_helpers

        if tool_requires:
            for reference in tool_requires:
                conanfile.requires.tool_require(repr(reference))
        if requires:
            for reference in requires:
                conanfile.requires(repr(reference))

        if python_requires:
            conanfile.python_requires = [pr.repr_notime() for pr in python_requires]

        if self._pyreq_loader:
            self._pyreq_loader.load_py_requires(conanfile, self, graph_lock, remotes,
                                                update, check_updates)

        conanfile._conan_is_consumer = True
        conanfile.generators = []  # remove the default txt generator
        return conanfile


def _parse_module(conanfile_module, module_id):
    """ Parses a python in-memory module, to extract the classes, mainly the main
    class defining the Recipe, but also process possible existing generators
    @param conanfile_module: the module to be processed
    @return: the main ConanFile class from the module
    """
    result = None
    for name, attr in conanfile_module.__dict__.items():
        if (name.startswith("_") or not inspect.isclass(attr) or
                attr.__dict__.get("__module__") != module_id):
            continue

        if issubclass(attr, ConanFile) and attr != ConanFile:
            if result is None:
                result = attr
            else:
                raise ConanException("More than 1 conanfile in the file")

    if result is None:
        raise ConanException("No subclass of ConanFile")

    return result


_load_python_lock = Lock()  # Loading our Python files is not thread-safe (modifies sys)


def _parse_conanfile(conanfile_path):
    with _load_python_lock:
        module, module_id = _load_python_file(conanfile_path)
    try:
        conanfile = _parse_module(module, module_id)
        return module, conanfile
    except Exception as e:  # re-raise with file name
        raise ConanException("%s: %s" % (conanfile_path, str(e)))


def load_python_file(conan_file_path):
    """ From a given path, obtain the in memory python import module
    """
    with _load_python_lock:
        module, module_id = _load_python_file(conan_file_path)
    return module, module_id


def _load_python_file(conan_file_path):
    """ From a given path, obtain the in memory python import module
    """

    if not os.path.exists(conan_file_path):
        raise NotFoundException("%s not found!" % conan_file_path)

    def new_print(*args, **kwargs):  # Make sure that all user python files print() goes to stderr
        kwargs.setdefault("file", sys.stderr)
        print(*args, **kwargs)

    module_id = str(uuid.uuid1())
    current_dir = os.path.dirname(conan_file_path)
    sys.path.insert(0, current_dir)
    try:
        old_modules = list(sys.modules.keys())
        with chdir(current_dir):
            old_dont_write_bytecode = sys.dont_write_bytecode
            try:
                sys.dont_write_bytecode = True
                spec = imp_util.spec_from_file_location(module_id, conan_file_path)
                loaded = imp_util.module_from_spec(spec)
                spec.loader.exec_module(loaded)
                sys.dont_write_bytecode = old_dont_write_bytecode
            except ImportError:
                version_txt = _get_required_conan_version_without_loading(conan_file_path)
                if version_txt:
                    validate_conan_version(version_txt)
                raise

            required_conan_version = getattr(loaded, "required_conan_version", None)
            if required_conan_version:
                validate_conan_version(required_conan_version)

        # These lines are necessary, otherwise local conanfile imports with same name
        # collide, but no error, and overwrite other packages imports!!
        added_modules = set(sys.modules).difference(old_modules)
        for added in added_modules:
            module = sys.modules[added]
            if module:
                try:
                    try:
                        # Most modules will have __file__ != None
                        folder = os.path.dirname(module.__file__)
                    except (AttributeError, TypeError):
                        # But __file__ might not exist or equal None
                        # Like some builtins and Namespace packages py3
                        folder = module.__path__._path[0]
                except AttributeError:  # In case the module.__path__ doesn't exist
                    pass
                else:
                    if folder.startswith(current_dir):
                        module = sys.modules.pop(added)
                        module.print = new_print
                        sys.modules["%s.%s" % (module_id, added)] = module
    except ConanException:
        raise
    except Exception:
        trace = traceback.format_exc().split('\n')
        raise ConanException("Unable to load conanfile in %s\n%s" % (conan_file_path,
                                                                     '\n'.join(trace[3:])))
    finally:
        sys.path.pop(0)

    loaded.print = new_print
    return loaded, module_id


def _get_required_conan_version_without_loading(conan_file_path):
    # First, try to detect the required_conan_version in "text" mode
    # https://github.com/conan-io/conan/issues/11239
    contents = load(conan_file_path)

    txt_version = None

    try:
        found = re.search(r"(.*)required_conan_version\s*=\s*[\"'](.*)[\"']", contents)
        if found and "#" not in found.group(1):
            txt_version = found.group(2)
    except:  # noqa this should be solid, cannot fail
        pass

    return txt_version


class ConanFileTextLoader:
    """Parse a conanfile.txt file"""

    def __init__(self, input_text):
        # Prefer composition over inheritance, the __getattr__ was breaking things
        self._config_parser = TextINIParse(input_text,  ["requires", "generators", "options",
                                                         "imports", "tool_requires", "test_requires",
                                                         "layout"],
                                           strip_comments=True)

    @property
    def layout(self):
        """returns the declared layout"""
        tmp = [r.strip() for r in self._config_parser.layout.splitlines()]
        if len(tmp) > 1:
            raise ConanException("Only one layout can be declared in the [layout] section of "
                                 "the conanfile.txt")
        return tmp[0] if tmp else None

    @property
    def requirements(self):
        """returns a list of requires
        EX:  "OpenCV/2.4.10@phil/stable"
        """
        return [r.strip() for r in self._config_parser.requires.splitlines()]

    @property
    def tool_requirements(self):
        """returns a list of tool_requires
        EX:  "OpenCV/2.4.10@phil/stable"
        """

        return [r.strip() for r in self._config_parser.tool_requires.splitlines()]

    @property
    def test_requirements(self):
        """returns a list of test_requires
        EX:  "gtest/2.4.10@phil/stable"
        """

        return [r.strip() for r in self._config_parser.test_requires.splitlines()]

    @property
    def options(self):
        return self._config_parser.options

    @property
    def generators(self):
        return self._config_parser.generators.splitlines()



from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_pure_runtime_dep():
    c = TestClient()
    c.save({"tooldep/conanfile.py": GenConanfile("tooldep", "0.1").with_package_file("bin/mysh",
                                                                                     "myshared")
                                                                  .with_package_type("shared-library"),
            "tool/conanfile.py": GenConanfile("tool", "0.1").with_requirement("tooldep/0.1")
                                                            .with_package_file("bin/myexe", "exe"),
            "lib/conanfile.py": GenConanfile("lib", "0.1").with_requirement("tool/0.1",
                                                                            headers=False,
                                                                            libs=False, run=True),
            "consumer/conanfile.py": GenConanfile().with_settings("build_type", "os",
                                                                  "compiler", "arch")
                                                   .with_requires("lib/0.1")})

    c.run("create tooldep")
    c.run("create tool")
    c.run("create lib")
    c.run("install consumer -g CMakeDeps")



import textwrap

from conan.test.utils.tools import TestClient


def test_exit_with_code():
    base = textwrap.dedent("""
        import sys
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello0"
            version = "0.1"

            def build(self):
                sys.exit(34)
        """)

    client = TestClient(light=True)
    client.save({"conanfile.py": base})
    client.run("install .")
    error_code = client.run("build .", assert_error=True)
    assert error_code == 34
    assert "Exiting with code: 34" in client.out



import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


conanfile_parent = """
from conan import ConanFile

class parentLib(ConanFile):
    name = "parent"
    version = "1.0"

    def package_info(self):
        self.cpp_info.cxxflags.append("-myflag")
        self.buildenv_info.define("MyEnvVar", "MyEnvVarValue")
"""


conanfile = """
import os
from conan import ConanFile
from pathlib import Path

class AConan(ConanFile):
    name = "lib"
    version = "1.0"

    # To save the folders and check later if the folder is the same
    copy_build_folder = None
    copy_source_folder = None
    copy_package_folder = None

    counter_package_calls = 0

    no_copy_source = %(no_copy_source)s
    requires = "parent/1.0@conan/stable"
    running_local_command = %(local_command)s

    def source(self):
        assert(self.source_folder == os.getcwd())
        assert(isinstance(self.source_path, Path))
        assert(str(self.source_path) == self.source_folder)

        # Prevented to use them, it's dangerous, because the source is run only for the first
        # config, so only the first build_folder/package_folder would be modified
        assert(self.build_folder is None)
        assert(self.package_folder is None)

        assert(self.source_folder is not None)
        self.copy_source_folder = self.source_folder

    def build(self):
        assert(self.build_folder == os.getcwd())
        assert(isinstance(self.build_path, Path))
        assert(str(self.build_path) == self.build_folder)

        if self.no_copy_source:
            assert(self.copy_source_folder == self.source_folder)  # Only in install
        else:
            assert(self.source_folder == self.build_folder)

        assert(self.package_folder is not None)
        assert(isinstance(self.package_path, Path))
        assert(str(self.package_path) == self.package_folder)
        self.copy_build_folder = self.build_folder

    def package(self):
        assert(self.build_folder == os.getcwd())
        assert(isinstance(self.build_path, Path))
        assert(str(self.build_path) == self.build_folder)

        if self.no_copy_source:
            assert(self.copy_source_folder == self.source_folder)  # Only in install
        else:
            assert(self.source_folder == self.build_folder)

        self.copy_package_folder = self.package_folder

    def package_info(self):
        assert(self.package_folder == os.getcwd())
        assert(isinstance(self.package_path, Path))
        assert(str(self.package_path) == self.package_folder)
"""


class TestFoldersAccess:
    """"Tests the presence of self.source_folder, self.build_folder, self.package_folder
    in the conanfile methods. Also the availability of the self.deps_cpp_info, self.deps_user_info
    and self.deps_env_info."""

    @pytest.fixture(autouse=True)
    def setup(self):
        self.client = TestClient(light=True)
        self.client.save({"conanfile.py": conanfile_parent})
        self.client.run("export . --user=conan --channel=stable")

    def test_source_local_command(self):
        c1 = conanfile % {"no_copy_source": False,
                          "local_command": True}
        self.client.save({"conanfile.py": c1}, clean_first=True)
        self.client.run("source .")

        c1 = conanfile % {"no_copy_source": True,
                          "local_command": True}
        self.client.save({"conanfile.py": c1}, clean_first=True)
        self.client.run("source .")

    def test_deploy(self):
        c1 = conanfile % {"no_copy_source": False,
                          "local_command": False}
        self.client.save({"conanfile.py": c1}, clean_first=True)
        self.client.run("create . --user=user --channel=testing --build missing")
        self.client.run("install --requires=lib/1.0@user/testing")  # Checks deploy

    def test_full_install(self):
        c1 = conanfile % {"no_copy_source": False,
                          "local_command": False}
        self.client.save({"conanfile.py": c1}, clean_first=True)
        self.client.run("create . --user=conan --channel=stable --build='*'")

        c1 = conanfile % {"no_copy_source": True,
                          "local_command": False}
        self.client.save({"conanfile.py": c1}, clean_first=True)
        self.client.run("create . --user=conan --channel=stable --build='*'")

        c1 = conanfile % {"no_copy_source": False,
                          "local_command": False}
        self.client.save({"conanfile.py": c1}, clean_first=True)
        self.client.run("create . --user=conan --channel=stable --build='*'")


class TestRecipeFolder:
    recipe_conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import load
        import os
        class Pkg(ConanFile):
            exports = "file.txt"
            def init(self):
                r = load(self, os.path.join(self.recipe_folder, "file.txt"))
                self.output.info("INIT: {}".format(r))
            def set_name(self):
                r = load(self, os.path.join(self.recipe_folder, "file.txt"))
                self.output.info("SET_NAME: {}".format(r))
            def configure(self):
                r = load(self, os.path.join(self.recipe_folder, "file.txt"))
                self.output.info("CONFIGURE: {}".format(r))
            def requirements(self):
                r = load(self, os.path.join(self.recipe_folder, "file.txt"))
                self.output.info("REQUIREMENTS: {}".format(r))
            def package(self):
                r = load(self, os.path.join(self.recipe_folder, "file.txt"))
                self.output.info("PACKAGE: {}".format(r))
            def package_info(self):
                r = load(self, os.path.join(self.recipe_folder, "file.txt"))
                self.output.info("PACKAGE_INFO: {}".format(r))
        """)

    def test_recipe_folder(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": self.recipe_conanfile,
                     "file.txt": "MYFILE!"})
        client.run("export . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "INIT: MYFILE!" in client.out
        assert "SET_NAME: MYFILE!" in client.out
        client.save({}, clean_first=True)
        client.run("install --requires=pkg/0.1@user/testing --build='*'")
        assert "pkg/0.1@user/testing: INIT: MYFILE!" in client.out
        assert "SET_NAME" not in client.out
        assert "pkg/0.1@user/testing: CONFIGURE: MYFILE!" in client.out
        assert "pkg/0.1@user/testing: REQUIREMENTS: MYFILE!" in client.out
        assert "pkg/0.1@user/testing: PACKAGE: MYFILE!" in client.out
        assert "pkg/0.1@user/testing: PACKAGE_INFO: MYFILE!" in client.out

    def test_local_flow(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": self.recipe_conanfile,
                     "file.txt": "MYFILE!"})
        client.run("install .")
        assert "INIT: MYFILE!" in client.out
        assert "SET_NAME: MYFILE!" in client.out
        assert "conanfile.py: CONFIGURE: MYFILE!" in client.out
        assert "conanfile.py: REQUIREMENTS: MYFILE!" in client.out

    def test_editable(self):
        client = TestClient(light=True)
        client.save({"pkg/conanfile.py": self.recipe_conanfile,
                     "pkg/file.txt": "MYFILE!",
                     "consumer/conanfile.py":
                         GenConanfile().with_require("pkg/0.1@user/stable")})
        client.run("editable add pkg --name=pkg --version=0.1 --user=user --channel=stable")

        client.run("install consumer")
        client.assert_listed_require({"pkg/0.1@user/stable": "Editable"})
        assert "pkg/0.1@user/stable: INIT: MYFILE!" in client.out
        assert "pkg/0.1@user/stable: CONFIGURE: MYFILE!" in client.out
        assert "pkg/0.1@user/stable: REQUIREMENTS: MYFILE!" in client.out
        assert "pkg/0.1@user/stable: PACKAGE_INFO: MYFILE!" in client.out



import platform
import pytest
import os
import textwrap
import pathlib

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.tools.gnu.makedeps import CONAN_MAKEFILE_FILENAME


def test_make_dirs_with_abs_path():
    """
    MakeDeps should support absolute paths when cppinfp
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile

        class TestMakeDirsConan(ConanFile):
            name = "mylib"
            version = "0.1"

            def package_info(self):
                self.cpp_info.frameworkdirs = []
                libname = "mylib"
                fake_dir = os.path.join("/", "my_absoulte_path", "fake")
                include_dir = os.path.join(fake_dir, libname, "include")
                lib_dir = os.path.join(fake_dir, libname, "lib")
                lib_dir2 = os.path.join(self.package_folder, "lib2")
                self.cpp_info.includedirs = [include_dir]
                self.cpp_info.libdirs = [lib_dir, lib_dir2]
                self.cpp_info.set_property("my_prop", "my prop value")
                self.cpp_info.set_property("my_prop_with_newline", "my\\nprop")
        """)
    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("create .")
    client.run("install --requires=mylib/0.1@ -g MakeDeps")

    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    print(makefile_content)
    prefix = pathlib.Path(client.current_folder).drive if platform.system() == "Windows" else ""
    assert 'CONAN_NAME_MYLIB = mylib' in makefile_content
    assert 'CONAN_VERSION_MYLIB = 0.1' in makefile_content
    assert f'CONAN_LIB_DIRS_MYLIB = \\\n\t$(CONAN_LIB_DIR_FLAG){prefix}/my_absoulte_path/fake/mylib/lib \\\n\t$(CONAN_LIB_DIR_FLAG)$(CONAN_ROOT_MYLIB)/lib2' in makefile_content
    assert f'CONAN_INCLUDE_DIRS_MYLIB = $(CONAN_INCLUDE_DIR_FLAG){prefix}/my_absoulte_path/fake/mylib/include' in makefile_content
    assert 'CONAN_BIN_DIRS_MYLIB = $(CONAN_BIN_DIR_FLAG)$(CONAN_ROOT_MYLIB)/bin' in makefile_content
    assert 'CONAN_PROPERTY_MYLIB_MY_PROP = my prop value' in makefile_content
    assert 'CONAN_PROPERTY_MYLIB_MY_PROP_WITH_NEWLINE' not in makefile_content
    assert "WARN: Skipping propery 'my_prop_with_newline' because it contains newline" in client.stderr

    lines = makefile_content.splitlines()
    for line_no, line in enumerate(lines):
        includedir_pattern = "CONAN_INCLUDE_DIRS_MYLIB = $(CONAN_INCLUDE_DIR_FLAG)"
        if line.startswith(includedir_pattern):
            assert os.path.isabs(line[len(includedir_pattern):])
            assert line.endswith("include")
        elif line.startswith("\t$(CONAN_LIB_DIR_FLAG)") and 'my_absoulte_path' in line:
            assert os.path.isabs(line[len("\t$(CONAN_LIB_DIR_FLAG)"):-2])
            assert line.endswith("lib \\")
        elif line.startswith("\t$(CONAN_LIB_DIR_FLAG)") and line.endswith('lib2'):
            assert "\t$(CONAN_LIB_DIR_FLAG)$(CONAN_ROOT_MYLIB)/lib2" in line


def test_make_empty_dirs():
    """
    MakeDeps should support cppinfo empty dirs
    """
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile

        class TestMakeDepsConan(ConanFile):
            name = "mylib"
            version = "0.1"

            def package_info(self):
                self.cpp_info.includedirs = []
                self.cpp_info.libdirs = []
                self.cpp_info.bindirs = []
                self.cpp_info.libs = []
                self.cpp_info.frameworkdirs = []
        """)
    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("create .")
    client.run("install --requires=mylib/0.1@ -g MakeDeps")

    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    assert 'CONAN_ROOT_MYLIB' in makefile_content
    assert 'SYSROOT' not in makefile_content
    assert 'CONAN_INCLUDE_DIRS' not in makefile_content
    assert 'CONAN_LIB_DIRS' not in makefile_content
    assert 'CONAN_BIN_DIRS' not in makefile_content
    assert 'CONAN_LIBS' not in makefile_content
    assert 'CONAN_FRAMEWORK_DIRS' not in makefile_content
    assert 'CONAN_PROPERTY' not in makefile_content


def test_libs_and_system_libs():
    """
    MakeDeps should support cppinfo system_libs with regular libs
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import save
        import os

        class TestMakeDepsConan(ConanFile):
            name = "mylib"
            version = "0.1"

            def package(self):
                save(self, os.path.join(self.package_folder, "lib", "file"), "")

            def package_info(self):
                self.cpp_info.libs = ["mylib1", "mylib2"]
                self.cpp_info.system_libs = ["system_lib1", "system_lib2"]
        """)
    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("create .")
    client.run("install --requires=mylib/0.1@ -g MakeDeps")

    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    assert "CONAN_LIBS_MYLIB = \\\n\t$(CONAN_LIB_FLAG)mylib1 \\\n\t$(CONAN_LIB_FLAG)mylib2" in makefile_content
    assert "CONAN_SYSTEM_LIBS_MYLIB = \\\n\t$(CONAN_SYSTEM_LIB_FLAG)system_lib1 \\\n\t$(CONAN_SYSTEM_LIB_FLAG)system_lib2" in makefile_content
    assert "CONAN_LIBS = $(CONAN_LIBS_MYLIB)" in makefile_content
    assert "CONAN_SYSTEM_LIBS = $(CONAN_SYSTEM_LIBS_MYLIB)" in makefile_content


def test_multiple_include_and_lib_dirs():
    """
    MakeDeps should support cppinfo multiple include directories
    """
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import save
        import os

        class TestMakeDepsConan(ConanFile):
            def package(self):
                for p in ["inc1", "inc2", "inc3/foo", "lib1", "lib2"]:
                    save(self, os.path.join(self.package_folder, p, "file"), "")

            def package_info(self):
                self.cpp_info.includedirs = ["inc1", "inc2", "inc3/foo"]
                self.cpp_info.libdirs = ["lib1", "lib2"]
        """)
    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=pkg --version=0.1")
    client.run("install --requires=pkg/0.1@ -g MakeDeps")

    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    assert "CONAN_INCLUDE_DIRS_PKG = \\\n" \
           "\t$(CONAN_INCLUDE_DIR_FLAG)$(CONAN_ROOT_PKG)/inc1 \\\n" \
           "\t$(CONAN_INCLUDE_DIR_FLAG)$(CONAN_ROOT_PKG)/inc2 \\\n" \
           "\t$(CONAN_INCLUDE_DIR_FLAG)$(CONAN_ROOT_PKG)/inc3/foo\n" in makefile_content
    assert "CONAN_LIB_DIRS_PKG = \\\n" \
           "\t$(CONAN_LIB_DIR_FLAG)$(CONAN_ROOT_PKG)/lib1 \\\n" \
           "\t$(CONAN_LIB_DIR_FLAG)$(CONAN_ROOT_PKG)/lib2\n" in makefile_content
    assert "CONAN_INCLUDE_DIRS = $(CONAN_INCLUDE_DIRS_PKG)\n" in makefile_content
    assert "CONAN_LIB_DIRS = $(CONAN_LIB_DIRS_PKG)\n" in makefile_content


def test_make_with_public_deps_and_component_requires():
    """
    Testing a complex structure like:

    * lib/0.1
        - Components: "cmp1"
    * other/0.1
    * second/0.1
        - Requires: "lib/0.1"
        - Components: "mycomponent", "myfirstcomp"
            + "mycomponent" requires "lib::cmp1"
            + "myfirstcomp" requires "mycomponent"
    * third/0.1
        - Requires: "second/0.1", "other/0.1"
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Recipe(ConanFile):

            def package_info(self):
                self.cpp_info.components["cmp1"].libs = ["libcmp1"]
    """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=lib --version=0.1")
    client.save({"conanfile.py": GenConanfile("other", "0.1").with_package_file("file.h", "0.1")})
    client.run("create .")

    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestMakeDepsConan(ConanFile):
            requires = "lib/0.1"

            def package_info(self):
                self.cpp_info.components["mycomponent"].requires.append("lib::cmp1")
                self.cpp_info.components["myfirstcomp"].requires.append("mycomponent")
                self.cpp_info.components["myfirstcomp"].set_property("my_prop", "my prop value")
                self.cpp_info.components["myfirstcomp"].set_property("my_prop_with_newline", "my\\nprop")

        """)
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("create . --name=second --version=0.1")
    client.save({"conanfile.py": GenConanfile("third", "0.1").with_package_file("file.h", "0.1")
                                                             .with_require("second/0.1")
                                                             .with_require("other/0.1")},
                clean_first=True)
    client.run("create .")

    client2 = TestClient(cache_folder=client.cache_folder)
    conanfile = textwrap.dedent("""
        [requires]
        third/0.1

        [generators]
        MakeDeps
        """)
    client2.save({"conanfile.txt": conanfile})
    client2.run("install .")

    makefile_content = client2.load(CONAN_MAKEFILE_FILENAME)
    assert "CONAN_DEPS = \\\n" \
           "\tthird \\\n" \
           "\tsecond \\\n" \
           "\tlib \\\n" \
           "\tother\n" in makefile_content
    assert 'CONAN_REQUIRES_SECOND = \\\n' \
           '\t$(CONAN_REQUIRES_SECOND_MYCOMPONENT) \\\n' \
           '\t$(CONAN_REQUIRES_SECOND_MYFIRSTCOMP)\n' in makefile_content
    assert 'SYSROOT' not in makefile_content
    assert 'CONAN_REQUIRES_SECOND_MYFIRSTCOMP = mycomponent\n' in makefile_content
    assert 'CONAN_LIBS_LIB = $(CONAN_LIBS_LIB_CMP1)\n' in makefile_content

    assert 'CONAN_COMPONENTS_LIB = cmp1\n' in makefile_content
    assert 'CONAN_LIBS_LIB_CMP1 = $(CONAN_LIB_FLAG)libcmp1\n' in makefile_content
    assert 'CONAN_REQUIRES = $(CONAN_REQUIRES_SECOND)\n' in makefile_content
    assert 'CONAN_LIBS = $(CONAN_LIBS_LIB)\n' in makefile_content

    assert 'CONAN_PROPERTY_SECOND_MYFIRSTCOMP_MY_PROP = my prop value\n' in makefile_content
    assert 'CONAN_PROPERTY_SECOND_MYFIRSTCOMP_MY_PROP_WITH_NEWLINE' not in makefile_content
    assert "WARN: Skipping propery 'my_prop_with_newline' because it contains newline" in client2.stderr


def test_make_with_public_deps_and_component_requires_second():
    """
    Testing another complex structure like:

    * other/0.1
        - Components: "cmp1", "cmp2", "cmp3"
            + "cmp1" (it shouldn't be affected by "other")
            + "cmp3" (it shouldn't be affected by "other")
            + "cmp3" requires "cmp1"
    * pkg/0.1
        - Requires: "other/0.1" -> "other::cmp1"
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Recipe(ConanFile):

            def package_info(self):
                self.cpp_info.components["cmp1"].libs = ["other_cmp1"]
                self.cpp_info.components["cmp2"].libs = ["other_cmp2"]
                self.cpp_info.components["cmp3"].requires.append("cmp1")
    """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=other --version=1.0")

    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestMakeDepsConan(ConanFile):
            requires = "other/1.0"

            def package_info(self):
                self.cpp_info.requires = ["other::cmp1"]
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=pkg --version=0.1")

    client2 = TestClient(cache_folder=client.cache_folder)
    conanfile = textwrap.dedent("""
        [requires]
        pkg/0.1

        [generators]
        MakeDeps
        """)
    client2.save({"conanfile.txt": conanfile})
    client2.run("install .")
    make_content = client2.load(CONAN_MAKEFILE_FILENAME)
    assert 'CONAN_REQUIRES_PKG = other::cmp1\n' in make_content
    assert 'CONAN_REQUIRES_OTHER = $(CONAN_REQUIRES_OTHER_CMP3)\n' in make_content
    assert 'CONAN_REQUIRES_OTHER_CMP3 = cmp1\n' in make_content
    assert 'CONAN_COMPONENTS_OTHER = \\\n\tcmp1 \\\n\tcmp2 \\\n\tcmp3\n' in make_content
    assert 'CONAN_LIBS_OTHER = \\\n\t$(CONAN_LIBS_OTHER_CMP1) \\\n\t$(CONAN_LIBS_OTHER_CMP2)\n' in make_content
    assert 'CONAN_LIBS_OTHER_CMP1 = $(CONAN_LIB_FLAG)other_cmp1\n' in make_content
    assert 'CONAN_LIBS_OTHER_CMP2 = $(CONAN_LIB_FLAG)other_cmp2\n' in make_content
    assert 'CONAN_LIBS = $(CONAN_LIBS_OTHER)\n' in make_content
    assert 'CONAN_REQUIRES = \\\n\t$(CONAN_REQUIRES_PKG) \\\n\t$(CONAN_REQUIRES_OTHER)\n' in make_content


def test_makedeps_with_test_requires():
    """
    MakeDeps has to create any test requires to be declared on the recipe.
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class TestMakeDeps(ConanFile):
            def package_info(self):
                self.cpp_info.libs = [f"lib{self.name}"]
        """)

    client.save({"conanfile.py": conanfile})
    client.run("create . --name=app --version=1.0")
    client.run("create . --name=test --version=1.0")
    # Create library having build and test requires
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class HelloLib(ConanFile):
            def build_requirements(self):
                self.tool_requires('app/1.0')
                self.test_requires('test/1.0')
        """)
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("install . -g MakeDeps")
    assert 'CONAN_DEPS = test\n' in client.load(CONAN_MAKEFILE_FILENAME)
    assert 'app' not in client.load(CONAN_MAKEFILE_FILENAME)


def test_makedeps_with_editable_layout():
    """
    The MakeDeps should be supported with editable layour mode
    """
    client = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import save
        class Dep(ConanFile):
            name = "dep"
            version = "0.1"
            def layout(self):
                self.cpp.source.includedirs = ["include"]
            def package_info(self):
                self.cpp_info.libs = ["mylib"]
        """)
    client.save({"dep/conanfile.py": dep,
                 "dep/include/header.h": "",
                 "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1")})
    client.run("create dep")
    client.run("editable add dep")
    with client.chdir("pkg"):
        client.run("install . -g MakeDeps")
        makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
        assert 'CONAN_LIBS_DEP = $(CONAN_LIB_FLAG)mylib\n' in makefile_content
        assert 'CONAN_INCLUDE_DIRS_DEP = $(CONAN_INCLUDE_DIR_FLAG)$(CONAN_ROOT_DEP)/include\n' in makefile_content


def test_makedeps_tool_requires():
    """
    Testing if MakeDeps are created for tool requires
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestMakeDepsConan(ConanFile):

            def package_info(self):
                self.cpp_info.libs = ["libtool"]
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name tool --version 1.0")

    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestMakeDepsConan(ConanFile):

            def package_info(self):
                self.cpp_info.components["cmp1"].libs = ["other_cmp1"]
                self.cpp_info.components["cmp2"].libs = ["other_cmp2"]
                self.cpp_info.components["cmp3"].requires.append("cmp1")
        """)
    client.save({"conanfile.py": conanfile}, clean_first=True)
    client.run("create . --name other --version 1.0")

    conanfile = textwrap.dedent("""
        [tool_requires]
        tool/1.0
        other/1.0
        [generators]
        MakeDeps
        """)
    client.save({"conanfile.txt": conanfile}, clean_first=True)
    client.run("install . -pr:h default -pr:b default")
    make_content = client.load(CONAN_MAKEFILE_FILENAME)

    # TODO: MakeDeps should support tool_requires in the future
    assert "CONAN_LIBS_TOOL" not in make_content
    assert "CONAN_NAME_OTHER" not in make_content


@pytest.mark.parametrize("pattern, result, expected",
                         [("libs = []", False, 'SYSROOT'),
                          ("sysroot = ['/foo/bar/sysroot']", True, 'CONAN_SYSROOT_PACKAGE = /foo/bar/sysroot')
                          if platform.system() != "Windows" else
                          ("sysroot = ['C:/my_sysroot']", True, 'CONAN_SYSROOT_PACKAGE = C:/my_sysroot')])

def test_makefile_sysroot(pattern, result, expected):
    """
    The MakeDeps should not enforce sysroot in case not defined
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
            from conan import ConanFile
            class TestMakeDepsConan(ConanFile):
                def package_info(self):
                    self.cpp_info.{{ pattern }}
            """)
    client.save({"conanfile.py": conanfile.replace("{{ pattern }}", pattern)})
    client.run("create . --name package --version 0.1.0")
    client.run("install --requires=package/0.1.0 -pr:h default -pr:b default -g MakeDeps")
    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    assert (expected in makefile_content) == result


def test_makefile_reference():
    """
    The MakeDeps should generate the correct package reference as variable
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Testing(ConanFile):
                pass
            """)
    client.save({"conanfile.py": conanfile})

    # official packages <name>/<version>
    client.run("create . --name package --version 0.1.0")
    client.run("install --requires=package/0.1.0 -pr:h default -pr:b default -g MakeDeps")
    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    assert 'CONAN_REFERENCE_PACKAGE = package/0.1.0\n' in makefile_content

    # custom packages <name>/<version>@<user>/<channel>
    client.run("create . --name package --version 0.1.0 --user user --channel channel")
    client.run("install --requires=package/0.1.0@user/channel -pr:h default -pr:b default -g MakeDeps")
    makefile_content = client.load(CONAN_MAKEFILE_FILENAME)
    assert 'CONAN_REFERENCE_PACKAGE = package/0.1.0@user/channel\n' in makefile_content



import platform
import os
import shutil

from conan.tools.build import cmd_args_to_string
from conan.tools.env.environment import Environment
from conan.errors import ConanException


class PipEnv:

    def __init__(self, conanfile, folder=None, name=""):
        """
        :param conanfile: The current conanfile "self"
        :param folder: Optional folder, by default the "build_folder"
        :param name: Optional name for the virtualenv, by default "conan_pipenv"
        """
        self._conanfile = conanfile
        self.env_name = f"conan_pipenv{f'_{name}' if name else ''}"
        self._env_dir = os.path.abspath(os.path.join(folder or conanfile.build_folder,
                                                     self.env_name))
        bins = "Scripts" if platform.system() == "Windows" else "bin"
        self.bin_dir = os.path.join(self._env_dir, bins)
        pyexe = "python.exe" if platform.system() == "Windows" else "python"
        self._python_exe = os.path.join(self.bin_dir, pyexe)
        self._create_venv()

    def generate(self):
        """
        Create a conan environment to use the python venv in the next steps of the conanfile.
        """
        env = Environment()
        env.prepend_path("PATH", self.bin_dir)
        env.vars(self._conanfile).save_script(self.env_name)

    @staticmethod
    def _default_python():
        python = "python" if platform.system() == "Windows" else "python3"
        default_python = shutil.which(python)
        return os.path.realpath(default_python) if default_python else None

    def _create_venv(self):
        python_interpreter = self._conanfile.conf.get("tools.system.pipenv:python_interpreter")
        python_interpreter = python_interpreter or self._default_python()
        if not python_interpreter:
            raise ConanException("PipEnv could not find a Python executable path. Please, install "
                                 "Python system-wide or set the "
                                 "'tools.system.pipenv:python_interpreter' "
                                 "conf to the full path of a Python executable")

        try:
            self._conanfile.run(cmd_args_to_string([python_interpreter, '-m', 'venv',
                                                    self._env_dir]))
        except ConanException as e:
            raise ConanException(f"PipEnv could not create a Python virtual "
                                 f"environment using '{python_interpreter}': {e}")

    def install(self, packages, pip_args=None):
        """
        Will try to install the list of pip packages passed as a parameter.

        :param packages: try to install the list of pip packages passed as a parameter.
        :param pip_args: additional argument list to be passed to the 'pip install' command,
                         e.g.: ['--no-cache-dir', '--index-url', 'https://my.pypi.org/simple'].
                         Defaults to ``None``.
        :return: the return code of the executed pip command.
        """
        args = [self._python_exe, "-m", "pip", "install", "--disable-pip-version-check"]
        if pip_args:
            args += list(pip_args)
        args += list(packages)
        command = cmd_args_to_string(args)
        return self._conanfile.run(command)



from collections import OrderedDict

from conan.tools.env.environment import ProfileEnvironment
from conan.internal.model.profile import Profile


class TestProfile:

    def test_profile_settings_update(self):
        new_profile = Profile()
        new_profile.update_settings(OrderedDict([("os", "Windows")]))

        new_profile.update_settings(OrderedDict([("OTHER", "2")]))
        assert new_profile.settings == OrderedDict([("os", "Windows"), ("OTHER", "2")])

        new_profile.update_settings(OrderedDict([("compiler", "2"), ("compiler.version", "3")]))
        assert new_profile.settings == OrderedDict([("os", "Windows"), ("OTHER", "2"),
                                      ("compiler", "2"), ("compiler.version", "3")])

    def test_profile_subsettings_update(self):
        new_profile = Profile()
        new_profile.update_settings(OrderedDict([("os", "Windows"),
                                                ("compiler", "Visual Studio"),
                                                ("compiler.runtime", "MT")]))

        new_profile.update_settings(OrderedDict([("compiler", "gcc")]))
        assert dict(new_profile.settings) == {"compiler": "gcc", "os": "Windows"}

        new_profile = Profile()
        new_profile.update_settings(OrderedDict([("os", "Windows"),
                                                 ("compiler", "Visual Studio"),
                                                 ("compiler.runtime", "MT")]))

        new_profile.update_settings(OrderedDict([("compiler", "Visual Studio"),
                                                 ("compiler.subsetting", "3"),
                                                 ("other", "value")]))

        assert dict(new_profile.settings) == {"compiler": "Visual Studio",
                                                      "os": "Windows",
                                                      "compiler.runtime": "MT",
                                                      "compiler.subsetting": "3",
                                                      "other": "value"}

    def test_package_settings_update(self):
        np = Profile()
        np.update_package_settings({"MyPackage": [("os", "Windows")]})

        np.update_package_settings({"MyPackage": [("OTHER", "2")]})
        assert np.package_settings_values == {"MyPackage": [("os", "Windows"), ("OTHER", "2")]}

        np._package_settings_values = None  # invalidate caching
        np.update_package_settings({"MyPackage": [("compiler", "2"), ("compiler.version", "3")]})
        assert np.package_settings_values == {"MyPackage": [("os", "Windows"), ("OTHER", "2"),
                                        ("compiler", "2"), ("compiler.version", "3")]}

    def test_profile_dump_order(self):
        # Settings
        profile = Profile()
        profile.package_settings["zlib"] = {"compiler": "gcc"}
        profile.settings["arch"] = "x86_64"
        profile.settings["compiler"] = "Visual Studio"
        profile.settings["compiler.version"] = "12"
        profile.tool_requires["*"] = ["zlib/1.2.8@lasote/testing"]
        profile.tool_requires["zlib/*"] = ["aaaa/1.2.3@lasote/testing",
                                                 "bb/1.2@lasote/testing"]
        assert """[settings]
arch=x86_64
compiler=Visual Studio
compiler.version=12
zlib:compiler=gcc
[tool_requires]
*: zlib/1.2.8@lasote/testing
zlib/*: aaaa/1.2.3@lasote/testing, bb/1.2@lasote/testing
""".splitlines() == profile.dumps().splitlines()

    def test_apply(self):
        # Settings
        profile = Profile()
        profile.settings["arch"] = "x86_64"
        profile.settings["compiler"] = "Visual Studio"
        profile.settings["compiler.version"] = "12"

        profile.update_settings(OrderedDict([("compiler.version", "14")]))

        assert '[settings]\narch=x86_64\ncompiler=Visual Studio\ncompiler.version=14\n' == profile.dumps()


def test_update_build_requires():
    # https://github.com/conan-io/conan/issues/8205#issuecomment-775032229
    profile = Profile()
    profile.tool_requires["*"] = ["zlib/1.2.8"]

    profile2 = Profile()
    profile2.tool_requires["*"] = ["zlib/1.2.8"]

    profile.compose_profile(profile2)
    assert profile.tool_requires["*"] == ["zlib/1.2.8"]

    profile3 = Profile()
    profile3.tool_requires["*"] = ["zlib/1.2.11"]

    profile.compose_profile(profile3)
    assert profile.tool_requires["*"] == ["zlib/1.2.11"]

    profile4 = Profile()
    profile4.tool_requires["*"] = ["cmake/2.7"]

    profile.compose_profile(profile4)
    assert profile.tool_requires["*"] == ["zlib/1.2.11", "cmake/2.7"]


def test_profile_serialize():
    profile = Profile()
    profile.options.loads("zlib*:aoption=1\nzlib*:otheroption=1")
    profile.buildenv = ProfileEnvironment.loads("VAR1=1\nVAR2=2")
    profile.conf.update("user.myfield:value", "MyVal")
    profile.settings["arch"] = "x86_64"
    profile.settings["compiler"] = "Visual Studio"
    profile.settings["compiler.version"] = "12"
    profile.tool_requires["*"] = ["zlib/1.2.8"]
    profile.update_package_settings({"MyPackage": [("os", "Windows")]})
    expected_json = {
        "settings": {"arch": "x86_64", "compiler": "Visual Studio", "compiler.version": "12"},
        "package_settings": {"MyPackage": {"os": "Windows"}}, "options": {},
        "tool_requires": {"*": ["'zlib/1.2.8'"]}, "conf": {"user.myfield:value": "MyVal"},
        "build_env": "VAR1=1\nVAR2=2\n"}
    assert expected_json == profile.serialize()

    profile.replace_requires.update({"cmake/*": "cmake/3.29.0"})
    profile.platform_tool_requires = ["cmake/3.29.0"]
    expected_json = {
        "settings": {"arch": "x86_64", "compiler": "Visual Studio", "compiler.version": "12"},
        "package_settings": {"MyPackage": {"os": "Windows"}}, "options": {},
        "replace_requires": {"cmake/*": "cmake/3.29.0"},
        "platform_tool_requires": ["cmake/3.29.0"],
        "tool_requires": {"*": ["'zlib/1.2.8'"]}, "conf": {"user.myfield:value": "MyVal"},
        "build_env": "VAR1=1\nVAR2=2\n"}
    assert expected_json == profile.serialize()



import textwrap
import pytest

from conan.internal.default_settings import default_settings_yml
from conan.errors import ConanException
from conan.internal.model.settings import Settings, bad_value_msg, undefined_field


def undefined_value(v):
    return "'%s' value not defined" % v


class TestSettingsLoads:

    def test_none_value(self):
        yml = "os: [null, Windows]"
        settings = Settings.loads(yml)
        settings.validate()
        assert settings.os == None
        assert "" == settings.dumps()
        settings.os = "Windows"
        assert settings.os == "Windows"
        assert "os=Windows" == settings.dumps()

    def test_any(self):
        yml = "os: [ANY]"
        settings = Settings.loads(yml)
        with pytest.raises(ConanException, match="'settings.os' value not defined"):
            settings.validate()  # Raise exception if unset
        settings.os = "some-os"
        settings.validate()
        assert settings.os == "some-os"
        assert "os=some-os" in settings.dumps()
        settings.os = "Windows"
        assert settings.os == "Windows"
        assert "os=Windows" == settings.dumps()

    def test_none_any(self):
        yml = "os: [null, ANY]"
        settings = Settings.loads(yml)
        settings.validate()
        settings.os = "None"
        settings.validate()
        assert settings.os == "None"
        assert "os=None" in settings.dumps()
        settings.os = "Windows"
        assert settings.os == "Windows"
        assert "os=Windows" == settings.dumps()

    def test_nested_any(self):
        yml = textwrap.dedent("""\
            os:
                ANY:
                    version: [null, ANY]
                Ubuntu:
                    version: ["18.04", "20.04"]
            """)
        settings = Settings.loads(yml)
        settings.os = "Windows"
        settings.validate()
        assert settings.os == "Windows"
        assert "os=Windows" in settings.dumps()
        settings.os.version = 2
        assert settings.os == "Windows"
        assert "os=Windows\nos.version=2" == settings.dumps()
        settings.os = "Ubuntu"
        with pytest.raises(ConanException, match="'settings.os.version' value not defined"):
            settings.validate()
        with pytest.raises(ConanException, match="Invalid setting '3' is not a valid 'settings.os.version'"):
            settings.os.version = 3
        settings.os.version = "20.04"
        assert "os=Ubuntu\nos.version=20.04" == settings.dumps()
        assert settings.os.version == "20.04"

    def test_getattr_none(self):
        yml = "os: [None, Windows]"
        settings = Settings.loads(yml)
        assert settings.os == None
        _os = getattr(settings, "os")
        assert _os == None
        assert str(_os) == "None"

    def test_get_safe(self):
        yml = "os: [None, Windows]"
        settings = Settings.loads(yml)
        settings.os = "Windows"
        assert settings.os == "Windows"
        assert settings.get_safe("compiler.version") == None
        assert settings.get_safe("build_type") == None
        assert "Release" == settings.get_safe("build_type", "Release")
        assert False == settings.get_safe("build_type", False)
        assert "Windows" == settings.get_safe("os", "Linux")

    def test_none_subsetting(self):
        yml = """os:
    null:
    Windows:
        subsystem: [null, cygwin]
"""
        settings = Settings.loads(yml)
        settings.validate()
        assert settings.os == None
        assert "" == settings.dumps()
        settings.os = "Windows"
        assert settings.os.subsystem == None
        assert "os=Windows" == settings.dumps()
        settings.os.subsystem = "cygwin"
        settings.validate()
        assert "os=Windows\nos.subsystem=cygwin" == settings.dumps()

    def test_none__sub_subsetting(self):
        yml = """os:
    null:
        subsystem: [null, cygwin]
    Windows:
"""
        with pytest.raises(ConanException, match="settings.yml: null setting can't have subsettings"):
            Settings.loads(yml)


class TestSettings:

    @pytest.fixture(autouse=True)
    def setup(self):
        data = {"compiler": {
                            "Visual Studio": {
                                             "version": ["10", "11", "12"],
                                             "runtime": ["MD", "MT"]},
                            "gcc": {
                                   "version": ["4.8", "4.9"],
                                   "arch": {"x86": {"speed": ["A", "B"]},
                                            "x64": {"speed": ["C", "D"]}}}
                                   },
                "os": ["Windows", "Linux"]}
        self.sut = Settings(data)

    def test_in_contains(self):
        self.sut.compiler = "Visual Studio"
        assert "Visual" in self.sut.compiler
        assert not ("Visual" not in self.sut.compiler)

    def test_os_split(self):
        settings = Settings.loads("""os:
    Windows:
    Linux:
    Macos:
        version: [1, 2]
    Android:
""")
        other_settings = Settings.loads("os: [Windows, Linux]")
        settings.os = "Windows"
        other_settings.os = "Windows"
        assert settings.dumps() == other_settings.dumps()

    def test_any(self):
        data = {"target": ["ANY"]}
        sut = Settings(data)
        sut.target = "native"
        assert sut.target == "native"

    def test_multi_os(self):
        settings = Settings.loads("""os:
            Windows:
            Linux:
                distro: [RH6, RH7]
            Macos:
                codename: [Mavericks, Yosemite]
        """)
        settings.os = "Windows"
        assert settings.os == "Windows"
        settings.os = "Linux"
        settings.os.distro = "RH6"
        assert settings.os.distro == "RH6"
        with pytest.raises(ConanException):
            settings.os.distro = "Other"
        with pytest.raises(ConanException):
            settings.os.codename = "Yosemite"
        settings.os = "Macos"
        settings.os.codename = "Yosemite"
        assert settings.os.codename == "Yosemite"

    def test_remove(self):
        del self.sut.compiler
        self.sut.os = "Windows"
        self.sut.validate()
        assert self.sut.dumps() == "os=Windows"

    def test_loads_default(self):
        settings = Settings.loads("""os: [Windows, Linux, Macos, Android, FreeBSD, SunOS]
arch: [x86, x86_64, arm]
compiler:
    sun-cc:
        version: ["5.10", "5.11", "5.12", "5.13", "5.14"]
    gcc:
        version: ["4.8", "4.9", "5.0"]
    msvc:
        runtime: [None, MD, MT, MTd, MDd]
        version: ["10", "11", "12"]
    clang:
        version: ["3.5", "3.6", "3.7"]

build_type: [None, Debug, Release]""")
        settings.compiler = "clang"
        settings.compiler.version = "3.5"
        assert settings.compiler == "clang"
        assert settings.compiler.version == "3.5"

    def test_loads(self):
        settings = Settings.loads("""
compiler:
    msvc:
        runtime: [MD, MT]
        version:
            '10':
                arch: ["32"]
            '11':
                &id1
                arch: ["32", "64"]
            '12':
                *id1
    gcc:
        arch:
            x64:
                speed: [C, D]
            x86:
                speed: [A, B]
        version: ['4.8', '4.9']
os: [Windows, Linux]
""")
        settings.update_values([('compiler', 'msvc'),
                                ('compiler.version', '10'),
                                ('compiler.version.arch', '32')])
        assert settings.values_list == [('compiler', 'msvc'),
                          ('compiler.version', '10'),
                          ('compiler.version.arch', '32')]

        settings.compiler.version = "10"
        settings.compiler.version.arch = "32"
        settings.compiler.version = "11"
        settings.compiler.version.arch = "64"
        settings.compiler.version = "12"
        settings.compiler.version.arch = "64"

        assert settings.values_list == [('compiler', 'msvc'),
                          ('compiler.version', '12'),
                          ('compiler.version.arch', '64')]

    def test_set_value(self):
        self.sut.update_values([("compiler", "Visual Studio")])
        assert self.sut.compiler == "Visual Studio"
        self.sut.update_values([("compiler.version", "12")])
        assert self.sut.compiler.version == "12"
        self.sut.update_values([("compiler", "gcc")])
        assert self.sut.compiler == "gcc"
        self.sut.update_values([("compiler.version", "4.8")])
        assert self.sut.compiler.version == "4.8"
        self.sut.update_values([("compiler.arch", "x86")])
        assert self.sut.compiler.arch == "x86"
        self.sut.update_values([("compiler.arch.speed", "A")])
        assert self.sut.compiler.arch.speed == "A"

    def test_constraint(self):
        s2 = ["os"]
        self.sut.constrained(s2)
        with pytest.raises(ConanException) as cm:
            self.sut.compiler
        assert str(cm.value) == str(undefined_field("settings", "compiler", ["os"], "settings"))
        self.sut.os = "Windows"
        self.sut.os = "Linux"

    def test_constraint2(self):
        s2 = ["os2"]
        with pytest.raises(ConanException) as cm:
            self.sut.constrained(s2)
        assert str(cm.value) == str(undefined_field("settings", "os2", ["compiler", "os"], "settings"))

    def test_constraint6(self):
        s2 = {"os", "compiler"}
        self.sut.constrained(s2)
        self.sut.compiler = "Visual Studio"
        with pytest.raises(ConanException) as cm:
            self.sut.compiler.arch
        assert str(cm.value) == str(undefined_field("settings.compiler", "arch",
                                                                ['runtime', 'version'], "Visual Studio"))
        self.sut.os = "Windows"
        self.sut.compiler.version = "11"
        self.sut.compiler.version = "12"

    def test_validate(self):
        with pytest.raises(ConanException, match=str(undefined_value("settings.compiler"))):
            self.sut.validate()

        self.sut.compiler = "gcc"
        with pytest.raises(ConanException, match="value not defined"):
            self.sut.validate()

        self.sut.compiler.arch = "x86"
        with pytest.raises(ConanException, match="value not defined"):
            self.sut.validate()

        self.sut.compiler.arch.speed = "A"
        with pytest.raises(ConanException, match="value not defined"):
            self.sut.validate()

        self.sut.compiler.version = "4.8"
        with pytest.raises(ConanException, match="value not defined"):
            self.sut.validate()

        self.sut.os = "Windows"
        self.sut.validate()
        assert self.sut.values_list == [("compiler", "gcc"),
                                                ("compiler.arch", "x86"),
                                                ("compiler.arch.speed", "A"),
                                                ("compiler.version", "4.8"),
                                                ("os", "Windows")]

    def test_validate2(self):
        self.sut.os = "Windows"
        self.sut.compiler = "Visual Studio"
        with pytest.raises(ConanException, match="value not defined"):
            self.sut.validate()

        self.sut.compiler.runtime = "MD"
        with pytest.raises(ConanException, match="value not defined"):
            self.sut.validate()

        self.sut.compiler.version = "10"
        self.sut.validate()

        assert self.sut.values_list == [("compiler", "Visual Studio"),
                                                ("compiler.runtime", "MD"),
                                                ("compiler.version", "10"),
                                                ("os", "Windows")]

    def test_basic(self):
        s = Settings({"os": ["Windows", "Linux"]})
        s.os = "Windows"
        with pytest.raises(ConanException) as cm:
            self.sut.compiler = "kk"
        assert str(cm.value) == bad_value_msg("settings.compiler", "kk", ['Visual Studio', 'gcc'])

    def test_my(self):
        assert self.sut.compiler == None

        with pytest.raises(ConanException) as cm:
            self.sut.compiler = "kk"
        assert str(cm.value) == bad_value_msg("settings.compiler", "kk", ['Visual Studio', 'gcc'])

        self.sut.compiler = "Visual Studio"
        assert str(self.sut.compiler) == "Visual Studio"
        assert self.sut.compiler == "Visual Studio"

        with pytest.raises(ConanException) as cm:
            self.sut.compiler.kk
        assert str(cm.value) == str(undefined_field("settings.compiler", "kk", ['runtime', 'version'],
                                             "Visual Studio"))

        assert self.sut.compiler.version == None

        with pytest.raises(ConanException) as cm:
            self.sut.compiler.version = "123"
        assert str(cm.value) == bad_value_msg("settings.compiler.version", "123", ['10', '11', '12'])

        self.sut.compiler.version = "12"
        assert self.sut.compiler.version == "12"
        assert str(self.sut.compiler.version) == "12"

        with pytest.raises(ConanException) as cm:
            assert self.sut.compiler == "kk"
        assert str(cm.value) == bad_value_msg("settings.compiler", "kk", ['Visual Studio', 'gcc'])

        assert not (self.sut.compiler == "gcc")
        assert self.sut.compiler == "Visual Studio"

        assert self.sut.compiler.version == "12"
        assert not (self.sut.compiler.version == "11")

        with pytest.raises(ConanException) as cm:
            assert self.sut.compiler.version == "13"
        assert str(cm.value) == bad_value_msg("settings.compiler.version", "13", ['10', '11', '12'])

        self.sut.compiler = "gcc"
        with pytest.raises(ConanException) as cm:
            self.sut.compiler.runtime
        assert str(cm.value) == str(undefined_field("settings.compiler", "runtime", "['arch', 'version']",
                                             "gcc"))

        self.sut.compiler.arch = "x86"
        self.sut.compiler.arch.speed = "A"
        assert self.sut.compiler.arch.speed == "A"

        with pytest.raises(ConanException) as cm:
            self.sut.compiler.arch.speed = "D"
        assert str(cm.value) == bad_value_msg("settings.compiler.arch.speed", "D", ['A', 'B'])

        self.sut.compiler.arch = "x64"
        self.sut.compiler.arch.speed = "C"
        assert self.sut.compiler.arch.speed == "C"

        with pytest.raises(ConanException) as cm:
            self.sut.compiler.arch.speed = "A"
        assert str(cm.value) == bad_value_msg("settings.compiler.arch.speed", "A", ['C', 'D'])

        self.sut.compiler.arch.speed = "D"
        assert self.sut.compiler.arch.speed == "D"


def test_possible_values():
    settings = Settings.loads(default_settings_yml)
    settings.compiler = "gcc"
    sot = settings.compiler.cppstd.possible_values()
    assert sot == [None, '98', 'gnu98', '11', 'gnu11', '14', 'gnu14', '17', 'gnu17', '20',
                   'gnu20', '23', 'gnu23', '26', 'gnu26']

    # We cannot access the child definition of a non declared setting
    with pytest.raises(Exception) as e:
        settings.os.version.possible_values()

    assert "'settings.os' value not defined" in str(e.value)

    # But if we have it declared, we can
    settings.os = "watchOS"
    sot = settings.os.version.possible_values()
    assert "4.0" in sot

    # We can get the whole settings definition and explore the dict
    sot = settings.possible_values()
    assert [None, "cygwin", "msys", "msys2", "wsl"] == sot["os"]["Windows"]["subsystem"]


def test_rm_safe():
    settings = Settings.loads(default_settings_yml)
    settings.rm_safe("compiler.cppstd")
    settings.rm_safe("compiler.libcxx")
    settings.compiler = "gcc"
    with pytest.raises(Exception) as e:
        settings.compiler.cppstd = "14"
    assert "'settings.compiler.cppstd' doesn't exist for 'gcc'" in str(e.value)
    with pytest.raises(Exception) as e:
        settings.compiler.libcxx = "libstdc++"
    assert "'settings.compiler.libcxx' doesn't exist for 'gcc'" in str(e.value)

    settings.compiler = "clang"
    with pytest.raises(Exception) as e:
        settings.compiler.cppstd = "14"
    assert "'settings.compiler.cppstd' doesn't exist for 'clang'" in str(e.value)
    with pytest.raises(Exception) as e:
        settings.compiler.libcxx = "libstdc++"
    assert "'settings.compiler.libcxx' doesn't exist for 'clang'" in str(e.value)


def test_rm_safe_wildcard():
    settings = Settings.loads(default_settings_yml)
    settings.compiler = "gcc"
    settings.compiler.version = "4.8"
    settings.compiler.libcxx = "libstdc++"
    settings.rm_safe("compiler.*")
    assert settings.compiler == "gcc"
    assert settings.get_safe("compiler.version") is None
    assert settings.get_safe("compiler.libcxx") is None


def test_settings_intel_cppstd_03():
    settings = Settings.loads(default_settings_yml)
    settings.compiler = "intel-cc"
    # This doesn't crash, it used to crash due to "03" not quoted in setting.yml
    settings.compiler.cppstd = "03"


def test_set_value_non_existing_values():
    data = {
        "compiler": {
            "gcc": {
                "version": ["4.8", "4.9"],
                "arch": {"x86": {"speed": ["A", "B"]},
                         "x64": {"speed": ["C", "D"]}}}
        },
        "os": ["Windows", "Linux"]
    }
    settings = Settings(data)
    with pytest.raises(ConanException) as cm:
        settings.update_values([("foo", "A")])
    assert "'settings.foo' doesn't exist for 'settings'" in str(cm.value)
    with pytest.raises(ConanException) as cm:
        settings.update_values([("foo.bar", "A")])
    assert "'settings.foo' doesn't exist for 'settings'" in str(cm.value)
    # it does not raise any error
    settings.update_values([("foo", "A")], raise_undefined=False)
    settings.update_values([("foo.bar", "A")], raise_undefined=False)



import os
import shutil

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestRemoveEditablePackageTest:

    @pytest.fixture()
    def client(self):
        t = TestClient()
        t.save({'conanfile.py': GenConanfile()})
        t.run('editable add . --name=lib --version=version --user=user --channel=name')
        t.run("editable list")
        assert "lib" in t.out
        return t

    def test_unlink(self, client):
        client.run('editable remove -r=lib/version@user/name')
        assert "Removed editable 'lib/version@user/name':" in client.out
        client.run("editable list")
        assert "lib" not in client.out

    def test_unlink_pattern(self, client):
        client.run('editable remove -r=*')
        assert "Removed editable 'lib/version@user/name':" in client.out
        client.run("editable list")
        assert "lib" not in client.out

    def test_remove_path(self, client):
        client.run("editable remove")
        assert "Removed editable 'lib/version@user/name':" in client.out
        client.run("editable list")
        assert "lib" not in client.out

    def test_unlink_not_linked(self, client):
        client.run('editable remove -r=otherlib/version@user/name')
        assert "WARN: No editables were removed" in client.out
        client.run("editable list")
        assert "lib" in client.out

    def test_removed_folder(self,):
        # https://github.com/conan-io/conan/issues/15038
        c = TestClient()
        c.save({'pkg/conanfile.py': GenConanfile()})
        c.run('editable add pkg --name=lib --version=version')
        shutil.rmtree(os.path.join(c.current_folder, "pkg"))
        # https://github.com/conan-io/conan/issues/16164
        # Making it possible, repeated issue
        c.run("editable remove pkg")
        assert "Removed editable 'lib/version'" in c.out
        c.run("editable list")
        assert "lib" not in c.out



import os

import pytest

from conan.internal.model.info import load_binary_info
from conan.api.model import RecipeReference
from conan.internal.paths import CONANFILE_TXT, CONANINFO
from conan.test.utils.tools import TestClient,  GenConanfile
from conan.internal.util.files import load


@pytest.fixture()
def client():
    c = TestClient()
    c.save_home({"settings.yaml": "os: [Windows, Macos, Linux, FreeBSD]\nos_build: [Windows, Macos]\narch_build: [x86_64]",
                 "profiles/default":  "[settings]\nos=Windows"})

    def base_conanfile(name):
        return GenConanfile(name, "0.1").with_option("language", [0, 1])\
            .with_default_option("language", 0).with_settings("os")

    c.save({"conanfile.py": base_conanfile("hello0")})
    c.run("export . --user=lasote --channel=stable")
    c.save({"conanfile.py": base_conanfile("hello1").with_requires("hello0/0.1@lasote/stable")})
    c.run("export . --user=lasote --channel=stable")
    c.save({"conanfile.py": base_conanfile("hello2").with_requires("hello1/0.1@lasote/stable")})
    c.run("export . --user=lasote --channel=stable")
    return c


def test_install_combined(client):
    client.run("install . --build=missing")
    client.run("install . --build=missing --build hello1/*")
    assert "hello0/0.1@lasote/stable: Already installed!" in client.out
    assert "hello1/0.1@lasote/stable: Forced build from source" in client.out


def test_install_transitive_cache(client):
    client.run("install --requires=hello2/0.1@lasote/stable --build=missing")
    assert "hello0/0.1@lasote/stable: Generating the package" in client.out
    assert "hello1/0.1@lasote/stable: Generating the package" in client.out
    assert "hello2/0.1@lasote/stable: Generating the package" in client.out


def test_upper_option(client):
    client.run("install conanfile.py -o hello2*:language=1 -o hello1*:language=0 "
               "-o hello0*:language=1 --build missing")
    package_id = client.created_package_id("hello0/0.1@lasote/stable")
    package_id2 = client.created_package_id("hello1/0.1@lasote/stable")
    ref = RecipeReference.loads("hello0/0.1@lasote/stable")
    pref = client.get_latest_package_reference(ref, package_id)
    hello0 = client.get_latest_pkg_layout(pref).package()

    hello0_info = os.path.join(hello0, CONANINFO)
    hello0_conan_info = load_binary_info(load(hello0_info))
    assert "1" == hello0_conan_info["options"]["language"]

    pref1 = client.get_latest_package_reference(RecipeReference.loads("hello1/0.1@lasote/stable"), package_id2)
    hello1 = client.get_latest_pkg_layout(pref1).package()
    hello1_info = os.path.join(hello1, CONANINFO)
    hello1_conan_info = load_binary_info(load(hello1_info))
    assert "0" == hello1_conan_info["options"]["language"]


def test_inverse_upper_option(client):
    client.run("install . -o language=0 -o hello1*:language=1 -o hello0*:language=0 --build missing")
    package_id = client.created_package_id("hello0/0.1@lasote/stable")
    package_id2 = client.created_package_id("hello1/0.1@lasote/stable")
    ref = RecipeReference.loads("hello0/0.1@lasote/stable")
    pref = client.get_latest_package_reference(ref, package_id)
    hello0 = client.get_latest_pkg_layout(pref).package()

    hello0_info = os.path.join(hello0, CONANINFO)
    hello0_conan_info = load_binary_info(load(hello0_info))
    assert "0" == hello0_conan_info["options"]["language"]

    pref1 = client.get_latest_package_reference(RecipeReference.loads("hello1/0.1@lasote/stable"), package_id2)
    hello1 = client.get_latest_pkg_layout(pref1).package()
    hello1_info = os.path.join(hello1, CONANINFO)
    hello1_conan_info = load_binary_info(load(hello1_info))
    assert "1" == hello1_conan_info["options"]["language"]


def test_upper_option_txt(client):
    files = {CONANFILE_TXT: """[requires]
        hello1/0.1@lasote/stable

        [options]
        hello0*:language=1
        hello1*:language=0
        """}
    client.save(files, clean_first=True)

    client.run("install . --build missing")
    package_id = client.created_package_id("hello0/0.1@lasote/stable")
    package_id2 = client.created_package_id("hello1/0.1@lasote/stable")
    ref = RecipeReference.loads("hello0/0.1@lasote/stable")
    pref = client.get_latest_package_reference(ref, package_id)
    hello0 = client.get_latest_pkg_layout(pref).package()
    hello0_info = os.path.join(hello0, CONANINFO)
    hello0_conan_info = load_binary_info(load(hello0_info))
    assert "1" == hello0_conan_info["options"]["language"]

    pref1 = client.get_latest_package_reference(RecipeReference.loads("hello1/0.1@lasote/stable"), package_id2)
    hello1 = client.get_latest_pkg_layout(pref1).package()
    hello1_info = os.path.join(hello1, CONANINFO)
    hello1_conan_info = load_binary_info(load(hello1_info))
    assert "0" == hello1_conan_info["options"]["language"]



import json

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


@pytest.mark.parametrize("config_version, mode, result", [
    ("myconfig/1.2.3#rev1", "minor_mode", "myconfig/1.2.Z"),
    ("myconfig/1.2.3#rev1", "patch_mode", "myconfig/1.2.3"),
    ("myconfig/1.2.3#rev1", "full_mode", "myconfig/1.2.3#rev1"),
    ("myconfig/1.2.3#rev1", "revision_mode", "myconfig/1.2.3#rev1"),
    ("myconfig/1.2.3", "minor_mode", "myconfig/1.2.Z")])
def test_config_package_id(config_version, mode, result):
    c = TestClient()
    config_version = json.dumps({"config_version": [config_version]})
    save(c.paths.config_version_path, config_version)
    save(c.paths.global_conf_path, f"core.package_id:config_mode={mode}")
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("create .")
    assert f"config_version: {result}" in c.out
    c.run("list pkg/0.1:* --format=json")
    info = json.loads(c.stdout)
    rrev = info["Local Cache"]["pkg/0.1"]["revisions"]["485dad6cb11e2fa99d9afbe44a57a164"]
    package_id = {"myconfig/1.2.Z": "c78b4d8224154390356fe04fe598d67aec930199",
                  "myconfig/1.2.3": "60005f5b11bef3ddd686b13f5c6bf576a9b882b8",
                  "myconfig/1.2.3#rev1:pid1": "3e4270809028e4566b55d3958e94ae3d6f0c92a7",
                  "myconfig/1.2.3#rev1": "aae875ae226416f177bf386a3e4ad6aaffce09e7"}
    package_id = package_id.get(result)
    pkg = rrev["packages"][package_id]
    assert pkg["info"] == {"config_version": [result]}


def test_error_config_package_id():
    c = TestClient()
    c.save_home({"global.conf": "core.package_id:config_mode=minor_mode"})
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("create .", assert_error=True)
    assert "ERROR: core.package_id:config_mode defined, " \
           "but error while loading 'config_version.json'" in c.out


@pytest.mark.parametrize("config_version, mode, result", [
    ("myconfig/1.2.3#rev1", "minor_mode", "myconfig/1.2.Z"),
    ("myconfig/1.2.3#rev1", "patch_mode", "myconfig/1.2.3"),
])
def test_config_package_id_clear(config_version, mode, result):
    c = TestClient(light=True)
    config_version = json.dumps({"config_version": [config_version]})
    save(c.paths.config_version_path, config_version)
    save(c.paths.global_conf_path, f"core.package_id:config_mode={mode}")
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_package_id("self.info.clear()")})
    c.run("create .")
    assert f"config_version: {result}" not in c.out


def test_recipe_revision_mode():
    clienta = TestClient()
    clienta.save_home({"global.conf": "core.package_id:default_unknown_mode=recipe_revision_mode"})

    clienta.save({"conanfile.py": GenConanfile()})
    clienta.run("create . --name=liba --version=0.1 --user=user --channel=testing")

    clientb = TestClient(cache_folder=clienta.cache_folder)
    clientb.save({"conanfile.py": GenConanfile("libb", "0.1").with_require("liba/0.1@user/testing")})
    clientb.run("create . --user=user --channel=testing")

    clientc = TestClient(cache_folder=clienta.cache_folder)
    clientc.save({"conanfile.py": GenConanfile("libc", "0.1").with_require("libb/0.1@user/testing")})
    clientc.run("install . --user=user --channel=testing")

    # Do a minor change to the recipe, it will change the recipe revision
    clienta.save({"conanfile.py": str(GenConanfile()) + "# comment"})
    clienta.run("create . --name=liba --version=0.1 --user=user --channel=testing")

    clientc.run("install . --user=user --channel=testing", assert_error=True)
    assert "ERROR: Missing prebuilt package for 'libb/0.1@user/testing'" in clientc.out
    # Building b with the new recipe revision of liba works
    clientc.run("install . --user=user --channel=testing --build=libb*")

    # Now change only the package revision of liba
    clienta.run("create . --name=liba --version=0.1 --user=user --channel=testing")
    clientc.run("install . --user=user --channel=testing")


def test_config_package_id_mode():
    # chicken and egg problem when the configuration package itself is affected by the mode
    c = TestClient(light=True)
    config_version = json.dumps({"config_version": ["myconfig/0.1"]})
    save(c.paths.config_version_path, config_version)
    c.save_home({"global.conf": "core.package_id:config_mode=minor_mode"})

    c.save({"conanfile.py": GenConanfile("myconfig", "0.1").with_package_type("configuration")})
    c.run("create . ")
    c.run("list *:*")
    # The binary is independent of the configuration version
    assert "config_version" not in c.out



import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


class TestBasic:

    def test_basic(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import CMakeToolchain
            class Pkg(ConanFile):
                def generate(self):
                    tc = CMakeToolchain(self)
                    tc.generate()
                """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("install .")

        assert "conanfile.py: Calling generate()" in client.out
        toolchain = client.load("conan_toolchain.cmake")
        assert "Conan automatically generated toolchain file" in toolchain

    def test_declarative(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                settings = "os", "compiler", "arch", "build_type"
                generators = ("CMakeToolchain", "CMakeDeps", "MesonToolchain")
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("install .")

        assert "conanfile.py: Generator 'CMakeToolchain' calling 'generate()'" in client.out
        assert "conanfile.py: Generator 'MesonToolchain' calling 'generate()'" in client.out
        assert "conanfile.py: Generator 'CMakeDeps' calling 'generate()'" in client.out
        toolchain = client.load("conan_toolchain.cmake")
        assert "Conan automatically generated toolchain file" in toolchain
        toolchain = client.load("conan_meson_native.ini")
        assert "[project options]" in toolchain

    @pytest.mark.tool("visual_studio")
    @pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
    def test_declarative_msbuildtoolchain(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                settings = "os", "compiler", "arch", "build_type"
                generators = ("MSBuildToolchain", )
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("install .")

        assert "conanfile.py: Generator 'MSBuildToolchain' calling 'generate()'" in client.out
        toolchain = client.load("conantoolchain.props")
        assert "<?xml version" in toolchain

    def test_error_missing_settings(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                generators = "MSBuildToolchain"
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("install .", assert_error=True)
        assert ("Error in generator 'MSBuildToolchain': 'settings.build_type' "
                "doesn't exist") in client.out

    def test_error_missing_settings_method(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.microsoft import MSBuildToolchain
            class Pkg(ConanFile):
                def generate(self):
                   tc = MSBuildToolchain(self)
                   tc.generate()
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("install .", assert_error=True)
        assert "ERROR: conanfile.py: Error in generate() method, line 6" in client.out

    def test_declarative_new_helper(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import CMake
            class Pkg(ConanFile):
                generators = "CMakeToolchain"
                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
            """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("build .", assert_error=True)  # No CMakeLists.txt
        assert '-DCMAKE_TOOLCHAIN_FILE="conan_toolchain.cmake"' in client.out
        assert "ERROR: conanfile.py: Error in build() method" in client.out

    @pytest.mark.tool("visual_studio")
    @pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
    def test_toolchain_windows(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.microsoft import MSBuildToolchain
            class Pkg(ConanFile):
                name = "Pkg"
                version = "0.1"
                settings = "os", "compiler", "arch", "build_type"
                generators = "MSBuildDeps"

                def generate(self):
                    tc = MSBuildToolchain(self)
                    tc.generate()
        """)

        client.save({"conanfile.py": conanfile})

        client.run('install . -s os=Windows -s compiler=msvc -s compiler.version=191'
                   ' -s compiler.runtime=dynamic')

        conan_toolchain_props = client.load("conantoolchain.props")
        assert "<ConanPackageName>Pkg</ConanPackageName>" in conan_toolchain_props
        assert "<ConanPackageVersion>0.1</ConanPackageVersion>" in conan_toolchain_props



import textwrap
import json

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save
import os

# Using the sbom tool with "conan create"
sbom_hook_post_package = """
import json
import os
from conan.errors import ConanException
from conan.api.output import ConanOutput
from conan.tools.sbom import {cyclone_version}

def post_package(conanfile):
    sbom_{cyclone_version} = {cyclone_version}(conanfile, add_build={add_build}, add_tests={add_tests})
    metadata_folder = conanfile.package_metadata_folder
    file_name = "sbom.cdx.json"
    with open(os.path.join(metadata_folder, file_name), 'w') as f:
        json.dump(sbom_{cyclone_version}, f, indent=4)
    ConanOutput().success(f"CYCLONEDX CREATED - {{conanfile.package_metadata_folder}}")
"""

# Using the sbom tool with "conan install"
sbom_hook_post_generate = """
import json
import os
from conan.errors import ConanException
from conan.api.output import ConanOutput
from conan.tools.sbom import {cyclone_version}

def post_generate(conanfile):
    sbom_{cyclone_version} = {cyclone_version}(conanfile, name={name})
    generators_folder = conanfile.generators_folder
    file_name = "sbom.cdx.json"
    os.mkdir(os.path.join(generators_folder, "sbom"))
    with open(os.path.join(generators_folder, "sbom", file_name), 'w') as f:
        json.dump(sbom_{cyclone_version}, f, indent=4)
    ConanOutput().success(f"CYCLONEDX CREATED - {{conanfile.generators_folder}}")
"""


@pytest.mark.parametrize("cyclone_version", ["cyclonedx_1_4", "cyclonedx_1_6"])
class TestCyclonedx:

    @pytest.fixture()
    def hook_setup_post_package(self, cyclone_version):
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_package.format(cyclone_version=cyclone_version,
                                                      add_build=True, add_tests=True))
        return tc


    @pytest.fixture()
    def hook_setup_post_package_no_tool_requires(self, cyclone_version):
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_package.format(cyclone_version=cyclone_version,
                                                      add_build=False, add_tests=True))
        return tc


    @pytest.fixture()
    def hook_setup_post_package_no_test(self, cyclone_version):
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_package.format(cyclone_version=cyclone_version,
                                                      add_build=True, add_tests=False))
        return tc


    @pytest.fixture()
    def hook_setup_post_package_tl(self, cyclone_version, transitive_libraries):
        tc = transitive_libraries
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_package.format(cyclone_version=cyclone_version,
                                                      add_build=True, add_tests=True))
        return tc


    def test_sbom_generation_create(self, hook_setup_post_package_tl):
        tc = hook_setup_post_package_tl
        tc.run("new cmake_lib -d name=bar -d version=1.0 -d requires=engine/1.0 -f")
        # bar -> engine/1.0 -> matrix/1.0
        tc.run("create . -tf=")
        bar_layout = tc.created_layout()
        assert os.path.exists(os.path.join(bar_layout.metadata(), "sbom.cdx.json"))


    def test_sbom_generation_skipped_dependencies(self, hook_setup_post_package):
        tc = hook_setup_post_package
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
                 "app/conanfile.py": GenConanfile("app", "1.0")
                                    .with_package_type("application")
                                    .with_requires("dep/1.0"),
                 "conanfile.py": GenConanfile("foo", "1.0").with_tool_requires("app/1.0")})
        tc.run("create dep")
        tc.run("create app")
        tc.run("create .")
        create_layout = tc.created_layout()

        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)
        # A skipped dependency also shows up in the sbom
        assert "pkg:conan/dep@1.0?rref=6a99f55e933fb6feeb96df134c33af44" in content

    @pytest.mark.parametrize("l, n", [('"simple"', 1), ('"multi1", "multi2"', 2), ('("tuple1", "tuple2")', 2)])
    def test_multi_license(self, hook_setup_post_package, l, n):
        tc = hook_setup_post_package
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            class HelloConan(ConanFile):
                name = 'foo'
                version = '1.0'
                license = {l}
        """)
        tc.save({"conanfile.py": conanfile})
        tc.run("create .")
        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = json.loads(tc.load(cyclone_path))
        assert len(content["components"][0]["licenses"]) == n

    @pytest.mark.parametrize("l, keys", [('"Mit"', ["id"]), ('"custom_license name"', ["name"]), ('("mIT", "custom")', ["id", "name"])])
    def test_license_spdx_valid(self, hook_setup_post_package, l, keys):
        tc = hook_setup_post_package
        conanfile = textwrap.dedent(f"""
                from conan import ConanFile
                class HelloConan(ConanFile):
                    name = 'foo'
                    version = '1.0'
                    license = {l}
            """)
        tc.save({"conanfile.py": conanfile})
        tc.run("create .")
        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = json.loads(tc.load(cyclone_path))
        for i, l in enumerate(content["components"][0]["licenses"]):
            assert next(iter(l["license"])) == keys[i]



    def test_sbom_generation_no_tool_requires(self, hook_setup_post_package_no_tool_requires):
        tc = hook_setup_post_package_no_tool_requires
        tc.save({"app/conanfile.py": GenConanfile("app", "1.0")
                                    .with_package_type("application"),
                 "conanfile.py": GenConanfile("foo", "1.0").with_tool_requires("app/1.0")})
        tc.run("create app")
        tc.run("create .")
        create_layout = tc.created_layout()

        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)

        assert "pkg:conan/app" not in content


    def test_sbom_generation_transitive_test_requires(self, hook_setup_post_package_no_test):
        tc = hook_setup_post_package_no_test
        tc.save({"test_re/conanfile.py": GenConanfile("test_re", "1.0"),
                 "app/conanfile.py": GenConanfile("app", "1.0")
                                    .with_package_type("application")
                                    .with_test_requires("test_re/1.0"),
                 "conanfile.py": GenConanfile("foo", "1.0").with_tool_requires("app/1.0")})
        tc.run("create test_re")

        tc.run("create app")
        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)
        assert "pkg:conan/test_re@1.0" not in content

        tc.run("create .")
        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)
        assert "pkg:conan/test_re@1.0" not in content


    def test_sbom_generation_dependency_test_require(self, hook_setup_post_package_no_test):
        tc = hook_setup_post_package_no_test
        tc.save({"special/conanfile.py": GenConanfile("special", "1.0"),
                 "foo/conanfile.py": GenConanfile("foo", "1.0")
                .with_test_requires("special/1.0"),
                 "conanfile.py": GenConanfile("bar", "1.0").with_tool_requires("foo/1.0").with_require("special/1.0")})
        tc.run("create special")
        tc.run("create foo")

        tc.run("create .")
        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)
        assert "pkg:conan/special@1.0" in content


    @pytest.fixture()
    def hook_setup_post_generate(self, cyclone_version):
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_generate.format(cyclone_version=cyclone_version, name=None))
        return tc


    def test_sbom_generation_install_requires(self, hook_setup_post_generate):
        tc = hook_setup_post_generate
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
                 "conanfile.py": GenConanfile("foo", "1.0").with_requires("dep/1.0")})
        tc.run("export dep")
        tc.run("create . --build=missing")

        # cli -> foo -> dep
        tc.run("install --requires=foo/1.0")
        assert os.path.exists(os.path.join(tc.current_folder, "sbom", "sbom.cdx.json"))


    def test_sbom_generation_install_path(self, hook_setup_post_generate):
        tc = hook_setup_post_generate
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
                 "conanfile.py": GenConanfile("foo", "1.0").with_requires("dep/1.0")})
        tc.run("create dep")

        # foo -> dep
        tc.run("install .")
        assert os.path.exists(os.path.join(tc.current_folder, "sbom", "sbom.cdx.json"))


    def test_sbom_generation_install_path_consumer(self, hook_setup_post_generate):
        tc = hook_setup_post_generate
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
                 "conanfile.py": GenConanfile().with_requires("dep/1.0")})
        tc.run("create dep")

        # conanfile.py -> dep
        tc.run("install .")
        assert os.path.exists(os.path.join(tc.current_folder, "sbom", "sbom.cdx.json"))


    def test_sbom_generation_install_path_txt(self, hook_setup_post_generate):
        tc = hook_setup_post_generate
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
                 "conanfile.txt": textwrap.dedent(
                     """
                     [requires]
                     dep/1.0
                     """
                 )})
        tc.run("create dep")

        # foo -> dep
        tc.run("install .")
        assert os.path.exists(os.path.join(tc.current_folder, "sbom", "sbom.cdx.json"))

    def test_sbom_with_special_root_node(self, hook_setup_post_generate):
        # In this test, we have only one node in the subgraph, which has build context, so the number
        # of components after processing it is zero.
        tc = hook_setup_post_generate
        package_name = "foo"

        conanfile =  textwrap.dedent("""
                from conan import ConanFile
                class FooPackage(ConanFile):
                    name = "foo"
                    version = "1.0"
                    package_type = "build-scripts"
        """)
        tc.save({"conanfile.py": conanfile})
        tc.run("create .")
        create_layout = tc.created_layout()
        assert os.path.exists(os.path.join(create_layout.build(), "sbom", "sbom.cdx.json"))
        with open(os.path.join(create_layout.build(), "sbom", "sbom.cdx.json"), 'r') as file:
            sbom_json = json.load(file)
            assert package_name in sbom_json["metadata"]["component"]["name"]

    @pytest.mark.parametrize("name, result", [
        ("None", "conan-sbom"),
        ('"custom-name"', "custom-name")
    ])
    def test_sbom_generation_custom_name(self, cyclone_version, name, result):
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_generate.format(cyclone_version=cyclone_version, name=name))

        tc.save({"conanfile.py": GenConanfile()})
        tc.run("install .")
        assert os.path.exists(os.path.join(tc.current_folder, "sbom", "sbom.cdx.json"))
        assert f'"name": "{result}"' in tc.load(os.path.join(tc.current_folder, "sbom", "sbom.cdx.json"))

    def test_cyclonedx_check_content(self, cyclone_version):
        _sbom_hook_post_package = textwrap.dedent("""
        import json
        import os
        from conan.errors import ConanException
        from conan.api.output import ConanOutput
        from conan.tools.sbom import {cyclone_version}

        def post_package(conanfile):
            sbom_cyclonedx= {cyclone_version}(conanfile)
            metadata_folder = conanfile.package_metadata_folder
            file_name = "sbom.cdx.json"
            with open(os.path.join(metadata_folder, file_name), 'w') as f:
                json.dump(sbom_cyclonedx, f, indent=4)
            ConanOutput().success(f"CYCLONEDX CREATED - {{conanfile.package_metadata_folder}}")
        """)
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, _sbom_hook_post_package.format(cyclone_version=cyclone_version))
        conanfile_bar = textwrap.dedent("""
                from conan import ConanFile
                class HelloConan(ConanFile):
                    name = 'bar'
                    version = '1.0'
                    author = 'conan-dev'
                    package_type = 'application'
            """)

        conanfile_foo = textwrap.dedent("""
            from conan import ConanFile
            class HelloConan(ConanFile):
                name = 'foo'
                version = '1.0'
                author = 'conan-dev'
                package_type = 'application'

                def requirements(self):
                    self.requires("bar/1.0")
        """)
        tc.save({"conanfile.py": conanfile_bar})
        tc.run("create .")
        tc.save({"conanfile.py": conanfile_foo})
        tc.run("create .")

        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)
        content_json = json.loads(content)
        if cyclone_version == 'cyclonedx_1_4':
            assert content_json["metadata"]["component"]["author"] == 'conan-dev'
            assert content_json["metadata"]["component"]["type"] == 'application'
            assert content_json["metadata"]["tools"][0]
            assert content_json["components"][0]["author"] == 'conan-dev'
            assert content_json["components"][0]["type"] == 'application'
        elif cyclone_version == 'cyclonedx_1_6':
            assert not content_json["metadata"]["component"].get("author")
            assert content_json["metadata"]["component"]["authors"][0]["name"] == 'conan-dev'
            assert content_json["metadata"]["component"]["type"] == 'application'
            assert content_json["metadata"]["tools"]["components"][0]
            assert not content_json["components"][0].get("author")
            assert content_json["components"][0]["authors"][0]["name"] == 'conan-dev'
            assert content_json["components"][0]["type"] == 'application'

    @pytest.mark.parametrize("user, channel, user_dep, channel_dep", [("user", None, "user_dep", None), ("user", "channel", "user_dep", "channel_dep")])
    def test_sbom_user_path(self, hook_setup_post_package_tl, user, channel, user_dep, channel_dep):
        tc = hook_setup_post_package_tl
        channel_ref = f"/{channel_dep}" if channel_dep else ""
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
                 "conanfile.py": GenConanfile("main", "1.0").with_requires(f"dep/1.0@{user_dep}{channel_ref}")})
        command = "create dep"
        if user: command += f" --user={user_dep}"
        if channel: command += f" --channel={channel_dep}"

        tc.run(command)

        command = "create ."
        if user: command += f" --user={user}"
        if channel: command += f" --channel={channel}"
        tc.run(command)

        create_layout = tc.created_layout()
        cyclone_path = os.path.join(create_layout.metadata(), "sbom.cdx.json")
        content = tc.load(cyclone_path)
        content_json = json.loads(content)

        assert content_json["components"][0]["bom-ref"].split("&user=")[1] == f"{user}&channel={channel}" if channel else user
        assert content_json["dependencies"][0]["dependsOn"][0].split("&user=")[1] == f"{user_dep}&channel={channel_dep}" if channel_dep else user_dep

    def test_sbom_test_requires_skipped(self, cyclone_version):
        tc = TestClient()
        hook_path = os.path.join(tc.paths.hooks_path, "hook_sbom.py")
        save(hook_path, sbom_hook_post_generate.format(cyclone_version=cyclone_version, name=None,
                                                       add_build=False, add_tests=False))
        tc.save({"dep/conanfile.py": GenConanfile("mydep", "1.0"),
                 "conanfile.py": GenConanfile("foo", "1.0").with_test_requires("mydep/1.0")})
        tc.run("create dep")
        tc.run("install .")
        content = tc.load("sbom/sbom.cdx.json")
        assert "mydep" not in content



from conan.internal.rest.rest_client_v2 import RestV2Methods
from unittest.mock import MagicMock, patch


@patch("conan.internal.rest.conan_requester.ConanRequester")
def test_check_credentials_set_token_to_unset_with_force_auth_true_and_token_none(
    conanRequesterMock,
):
    return_status = MagicMock()
    return_status.status_code = 200
    conanRequesterMock.get = MagicMock(return_value=return_status)

    rest_v2_methods = RestV2Methods("foo", None, conanRequesterMock, None, True)
    rest_v2_methods.check_credentials(True)

    assert conanRequesterMock.get.call_args[1]["auth"].bearer == "Bearer unset"


@patch("conan.internal.rest.conan_requester.ConanRequester")
def test_check_credentials_keep_token_none_with_force_auth_false_and_token_none(
    conanRequesterMock,
):
    return_status = MagicMock()
    return_status.status_code = 200
    conanRequesterMock.get = MagicMock(return_value=return_status)

    rest_v2_methods = RestV2Methods("foo", None, conanRequesterMock, None, True)
    rest_v2_methods.check_credentials(False)

    assert conanRequesterMock.get.call_args[1]["auth"].bearer is None



import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("typedep, typeconsumer, different_id",
                         [("header-library", "application", False),
                          ("header-library", "shared-library", False),
                          ("header-library", "static-library", False),
                          ("static-library", "application", True),
                          ("static-library", "shared-library", True),
                          ("static-library", "header-library", False),
                          ("static-library", "static-library", False),
                          ("shared-library", "header-library", False),
                          ("shared-library", "static-library", False),
                          ("shared-library", "shared-library", False),
                          ("shared-library", "application", False)
                          ])
def test_default_package_id_options(typedep, typeconsumer, different_id):
    """ test that some consumer package ids are changed when the dependency change one of its
    options
    """
    c = TestClient(light=True)
    dep = GenConanfile("dep", "0.1").with_option("myopt", [True, False]) \
        .with_package_type(typedep).with_class_attribute('implements = ["auto_shared_fpic", "auto_header_only"]')
    consumer = GenConanfile("consumer", "0.1").with_requires("dep/0.1")\
        .with_package_type(typeconsumer).with_class_attribute('implements = ["auto_shared_fpic", "auto_header_only"]')

    c.save({"dep/conanfile.py": dep,
            "consumer/conanfile.py": consumer})
    c.run("create dep -o dep/*:myopt=True")
    pid1 = c.created_package_id("dep/0.1")
    c.run("create dep -o dep/*:myopt=False")
    pid2 = c.created_package_id("dep/0.1")
    if typedep != "header-library":
        assert pid1 != pid2

    c.run("create consumer -o dep/*:myopt=True")
    pid1 = c.created_package_id("consumer/0.1")
    c.run("create consumer -o dep/*:myopt=False")
    pid2 = c.created_package_id("consumer/0.1")
    if different_id:
        assert pid1 != pid2
    else:
        assert pid1 == pid2


@pytest.mark.parametrize("typedep, versiondep, typeconsumer, different_id",
                         [("static-library", "1.1", "header-library", False),
                          ("static-library", "1.0.1", "static-library", False),
                          ("static-library", "1.1", "static-library", True),
                          ("shared-library", "1.1", "header-library", False),
                          ("shared-library", "1.0.1", "static-library", False),
                          ("shared-library", "1.1", "static-library", True),
                          ("shared-library", "1.0.1", "shared-library", False),
                          ("shared-library", "1.1", "shared-library", True),
                          ("shared-library", "1.0.1", "application", False),
                          ("shared-library", "1.1", "application", True),
                          ("application", "2.1", "application", False),
                          ])
def test_default_package_id_versions(typedep, versiondep, typeconsumer, different_id):
    """ test that some consumer package ids are changed when the dependency changes its version
    """
    c = TestClient(light=True)
    dep = GenConanfile("dep").with_package_type(typedep)
    consumer = GenConanfile("consumer", "0.1").with_requires("dep/[>0.0]") \
        .with_package_type(typeconsumer)
    c.save({"dep/conanfile.py": dep,
            "consumer/conanfile.py": consumer})
    c.run("create dep --version=1.0")
    c.run("create consumer")
    pid1 = c.created_package_id("consumer/0.1")

    c.run(f"create dep --version={versiondep}")
    c.run("create consumer")
    pid2 = c.created_package_id("consumer/0.1")
    if different_id:
        assert pid1 != pid2
    else:
        assert pid1 == pid2



import textwrap

import pytest

from conan.test.utils.tools import TestClient


class TestAuthRemotePlugin:

    def test_error_auth_remote_plugin(self):
        """ Test when the plugin fails, we want a clear message and a helpful trace
        """
        c = TestClient(default_server_user=True)
        auth_plugin = textwrap.dedent("""\
            def auth_remote_plugin(remote, user=None):
                raise Exception("Test Error")
            """)
        c.save_home({"extensions/plugins/auth_remote.py": auth_plugin})
        c.run("remote logout default")
        c.run("remote login default", assert_error=True)
        assert "Error while processing 'auth_remote.py' plugin" in c.out
        assert "ERROR: Error while processing 'auth_remote.py' plugin, line " in c.out

    @pytest.mark.parametrize("password", ["password", "bad-password"])
    def test_auth_remote_plugin_direct_credentials(self, password):
        """ Test when the plugin give a correct and wrong password, we want a message about
        the success or fail in login
        """
        should_fail = password == "bad-password"
        c = TestClient(default_server_user=True)
        auth_plugin = textwrap.dedent(f"""\
            def auth_remote_plugin(remote, user=None):
                return "admin", "{password}"
            """)
        c.save_home({"extensions/plugins/auth_remote.py": auth_plugin})
        c.run("remote logout default")
        c.run("remote login default", assert_error=should_fail)
        if should_fail:
            assert "ERROR: Wrong user or password. [Remote: default]" in c.out
        else:
            assert ("Changed user of remote 'default' from 'None' (anonymous) to "
                    "'admin' (authenticated)") in c.out

    def test_auth_remote_plugin_fallback(self):
        """ Test when the plugin do not give any user or password, we want the code to continue with
            the rest of the input methods
        """
        c = TestClient(default_server_user=True)
        auth_plugin = textwrap.dedent("""\
                def auth_remote_plugin(remote, user=None):
                    return None, None
                """)
        c.save_home({"extensions/plugins/auth_remote.py": auth_plugin})
        c.run("remote logout default")
        c.run("remote login default")
        # As the auth plugin is not returning any password the code is falling back to the rest of
        # the input methods in this case the stdin provided by TestClient.
        assert ("Changed user of remote 'default' from 'None' (anonymous) to "
                "'admin' (authenticated)") in c.out



import os
import textwrap

from conan.tools.build import load_toolchain_args, CONAN_TOOLCHAIN_ARGS_FILE
from conan.test.utils.tools import TestClient


def test_autotools_namespace():
    client = TestClient()
    namespace = "somename"
    conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.gnu import AutotoolsToolchain, Autotools

            class Conan(ConanFile):
                settings = "os", "arch", "compiler", "build_type"
                def generate(self):
                    autotools = AutotoolsToolchain(self, namespace='{0}')
                    autotools.configure_args = ['a', 'b']
                    autotools.make_args = ['c', 'd']
                    autotools.generate()
                def build(self):
                    autotools = Autotools(self, namespace='{0}')
                    self.output.info(autotools._configure_args)
                    self.output.info(autotools._make_args)
            """.format(namespace))

    client.save({"conanfile.py": conanfile})
    client.run("install .")
    assert os.path.isfile(os.path.join(client.current_folder,
                                       "{}_{}".format(namespace, CONAN_TOOLCHAIN_ARGS_FILE)))
    content = load_toolchain_args(generators_folder=client.current_folder, namespace=namespace)
    at_configure_args = content.get("configure_args")
    at_make_args = content.get("make_args")
    client.run("build .")
    assert at_configure_args in client.out
    assert at_make_args in client.out



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient

conanfile = """import os
from conan import ConanFile

class MyTest(ConanFile):
    name = "pkg"
    version = "0.1"
    settings = "os", "build_type"
    build_policy = "missing"

    def build_id(self):
        if self.settings.os == "Windows":
            self.info_build.settings.build_type = "Any"

    def build(self):
        self.output.info("Building my code!")

    def package(self):
        self.output.info("Packaging %s!" % self.settings.build_type)
"""


class TestBuildIdTest:

    def test_create(self):
        # Ensure that build_id() works when multiple create calls are made
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create .  -s os=Windows -s build_type=Release")
        assert "pkg/0.1: Calling build()" in client.out
        assert "Building my code!" in client.out
        assert "Packaging Release!" in client.out
        # Debug must not build
        client.run("create .  -s os=Windows -s build_type=Debug")
        assert "pkg/0.1: Calling build()" not in client.out
        assert "Building my code!" not in client.out
        assert "Packaging Debug!" in client.out

        client.run("create . -s os=Linux -s build_type=Release")
        assert "pkg/0.1: Calling build()" in client.out
        assert "Building my code!" in client.out
        assert "Packaging Release!" in client.out
        client.run("create .  -s os=Linux -s build_type=Debug")
        assert "Building my code!" in client.out
        assert "pkg/0.1: Calling build()" in client.out
        assert "Packaging Debug!" in client.out

    def test_basic_install(self):
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("export . ")
        # Windows Debug
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Debug')
        assert "Building my code!" in client.out
        assert "Packaging Debug!" in client.out

        # Package Windows Release, it will reuse the previous build
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Release')
        assert "Building my code!" not in client.out
        assert "Packaging Release!" in client.out

        # Now Linux Debug
        client.run('install --requires=pkg/0.1 -s os=Linux -s build_type=Debug')
        assert "Building my code!" in client.out
        assert "Packaging Debug!" in client.out

        # Linux Release must build again, as it is not affected by build_id()
        client.run('install --requires=pkg/0.1 -s os=Linux -s build_type=Release')
        assert "Building my code!" in client.out
        assert "Packaging Release!" in client.out

        # But if the packages are removed, and we change the order, keeps working
        client.run("remove *:* -c")
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Release')
        assert "Building my code!" in client.out
        assert "Packaging Release!" in client.out

        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Debug')
        assert "Building my code!" not in client.out
        assert "Packaging Debug!" in client.out

    def test_clean_build(self):
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("export . ")
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Debug')
        # Package Windows Release, it will reuse the previous build
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Release')
        assert "Building my code!" not in client.out
        assert "Packaging Release!" in client.out

        client.run("cache clean")  # packages are still there
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Debug')
        assert "Building my code!" not in client.out
        assert "Packaging Release!" not in client.out
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Release')
        assert "Building my code!" not in client.out
        assert "Packaging Release!" not in client.out

        # Lets force the first rebuild, different order Linux first
        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Release --build=pkg*')
        assert "Building my code!" in client.out
        assert "Packaging Release!" in client.out

        client.run('install --requires=pkg/0.1 -s os=Windows -s build_type=Debug --build=pkg*')
        assert "Building my code!" not in client.out
        assert "Packaging Debug!" in client.out

    def test_failed_build(self):
        # Repeated failed builds keep failing
        fail_conanfile = textwrap.dedent("""\
            from conan import ConanFile
            class MyTest(ConanFile):
                settings = "build_type"
                def build(self):
                    raise Exception("Failed build!!")
            """)
        client = TestClient()
        # NORMAL case, every create fails
        client.save({"conanfile.py": fail_conanfile})
        client.run("create . --name=pkg --version=0.1 ", assert_error=True)
        assert "ERROR: pkg/0.1: Error in build() method, line 5" in client.out
        client.run("create . --name=pkg --version=0.1 ", assert_error=True)
        assert "ERROR: pkg/0.1: Error in build() method, line 5" in client.out
        # now test with build_id
        client.save({"conanfile.py": fail_conanfile +
                     "    def build_id(self): self.info_build.settings.build_type = 'any'"})
        client.run("create . --name=pkg --version=0.1 ", assert_error=True)
        assert "ERROR: pkg/0.1: Error in build() method, line 5" in client.out
        client.run("create . --name=pkg --version=0.1 ", assert_error=True)
        assert "ERROR: pkg/0.1: Error in build() method, line 5" in client.out

    def test_any_os_arch(self):
        conanfile_os = textwrap.dedent("""
            import os
            from conan import ConanFile

            class MyTest(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "arch", "build_type", "compiler"

                def build_id(self):
                    self.info_build.settings.build_type = "AnyValue"
                    self.info_build.settings.os = "AnyValue"
                    self.info_build.settings.arch = "AnyValue"
                    self.info_build.settings.compiler = "AnyValue"

                def build(self):
                    self.output.info("Building my code!")

                def package(self):
                    self.output.info("Packaging %s-%s!" % (self.settings.os, self.settings.arch))
            """)

        client = TestClient()
        client.save({"conanfile.py": conanfile_os})
        client.run("create .  -s os=Windows -s arch=x86_64 -s build_type=Release")
        assert "pkg/0.1: Calling build()" in client.out
        assert "Building my code!" in client.out
        assert "Packaging Windows-x86_64!" in client.out
        # Others must not build
        client.run("create .  -s os=Linux -s arch=x86 -s build_type=Debug")
        assert "pkg/0.1: Calling build()" not in client.out
        assert "Building my code!" not in client.out
        assert "Packaging Linux-x86!" in client.out


def test_remove_require():
    c = TestClient()
    remove = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "consumer"
            version = "1.0"
            requires = "dep/1.0"
            def build_id(self):
                self.info_build.requires.remove("dep")
        """)
    c.save({"dep/conanfile.py": GenConanfile("dep", "1.0"),
            "consumer/conanfile.py": remove})
    c.run("create dep")
    c.run("create consumer")


def test_build_id_error():
    """
    https://github.com/conan-io/conan/issues/14537
    This was failing because the ``DB get_matching_build_id()`` function was returning all prefs
    for a given ref, and the order could be by alphabetic package_id, but only the first one really
    contains the build folder with the artifacts. It was fixed by defining the DB "build_id" only
    for the first reference containing that build_id, not all
    """
    c = TestClient()
    myconan = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.scm import Version

        class BuildIdTestConan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "arch", "compiler", "build_type"
            options = {"shared": [True, False]}
            default_options = {"shared": False}

            def build_id(self):
                self.info_build.settings.build_type = "Any"
                self.info_build.options.shared = "Any"

            def package_id(self):
                compilerVer = Version(str(self.info.settings.compiler.version))
                if self.info.settings.compiler == "gcc":
                    if compilerVer >= "7":
                        self.info.settings.compiler.version = "7+"
    """)
    host_profile = textwrap.dedent("""
        [settings]
        arch = x86_64
        build_type = Release
        compiler = gcc
        compiler.cppstd = gnu17
        compiler.libcxx = libstdc++11
        compiler.version = 8
        os = Linux
    """)

    c.save({"conanfile.py": myconan,
            "myprofile": host_profile})
    c.run("create . "
          "-pr:h myprofile -pr:b myprofile "
          "-s build_type=Debug -o pkg/*:shared=True")
    c.assert_listed_binary({"pkg/0.1": ("538f60f3919ea9b8ea9e7c63c5948abe44913bec", "Build")})
    assert "pkg/0.1: build_id() computed 7d169d0d018d239cff27eb081e3f6575554e05c5" in c.out
    c.run("create . "
          "-pr:h myprofile -pr:b myprofile "
          "-s build_type=Debug -o pkg/*:shared=False")
    c.assert_listed_binary({"pkg/0.1": ("ba41c80b0373ef66e2ca95ed56961153082fbcd9", "Build")})
    assert "pkg/0.1: build_id() computed 7d169d0d018d239cff27eb081e3f6575554e05c5" in c.out
    assert "pkg/0.1: Won't be built, using previous build folder as defined in build_id()" in c.out
    c.run("create . "
          "-pr:h myprofile -pr:b myprofile "
          "-s build_type=Release -o pkg/*:shared=True")
    c.assert_listed_binary({"pkg/0.1": ("2eb252449df37d245568e32e5a41d0540db3c6e2", "Build")})
    assert "pkg/0.1: build_id() computed 7d169d0d018d239cff27eb081e3f6575554e05c5" in c.out
    assert "pkg/0.1: Won't be built, using previous build folder as defined in build_id()" in c.out
    c.run("create . "
          "-pr:h myprofile -pr:b myprofile "
          "-s build_type=Release -o pkg/*:shared=False")
    c.assert_listed_binary({"pkg/0.1": ("978c5442906f96846ccb201129ba1559071ce4ab", "Build")})
    assert "pkg/0.1: build_id() computed 7d169d0d018d239cff27eb081e3f6575554e05c5" in c.out
    assert "pkg/0.1: Won't be built, using previous build folder as defined in build_id()" in c.out



from parameterized import parameterized

from conan.internal.api.detect.detect_api import _parse_gnu_libc, _parse_musl_libc


class TestDetectLibc:
    @parameterized.expand(
        [
            [
                """ldd (Debian GLIBC 2.36-9+rpt2+deb12u4) 2.36
Copyright (C) 2022 Free Software Foundation, Inc.
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
Written by Roland McGrath and Ulrich Drepper.""",
                "2.36",
            ],
            [
                """ldd (Ubuntu GLIBC 2.35-0ubuntu3.6) 2.35
Copyright (C) 2022 Free Software Foundation, Inc.
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
Written by Roland McGrath and Ulrich Drepper.""",
                "2.35",
            ],
            [
                """ldd (GNU libc) 2.38
Copyright (C) 2023 Free Software Foundation, Inc.
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
Written by Roland McGrath and Ulrich Drepper.""",
                "2.38",
            ],
            [
                """musl libc (x86_64)
Version 1.2.4_git20230717
Dynamic Program Loader
Usage: /lib/ld-musl-x86_64.so.1 [options] [--] pathname""",
                None,
            ],
        ]
    )
    def test_parse_gnu_libc(self, ldd_output, expected_glibc_version):
        parsed_glibc_version = _parse_gnu_libc(ldd_output)
        assert expected_glibc_version == parsed_glibc_version

    @parameterized.expand(
        [
            [
                """musl libc (x86_64)
Version 1.2.4
Dynamic Program Loader
Usage: /lib/ld-musl-x86_64.so.1 [options] [--] pathname""",
                "1.2.4",
            ],
            [
                """musl libc (x86_64)
Version 1.2.4_git20230717
Dynamic Program Loader
Usage: /lib/ld-musl-x86_64.so.1 [options] [--] pathname""",
                "1.2.4_git20230717",
            ],
            [
                """ldd (Debian GLIBC 2.36-9+rpt2+deb12u4) 2.36
Copyright (C) 2022 Free Software Foundation, Inc.
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
Written by Roland McGrath and Ulrich Drepper.""",
                None,
            ],
            [
                """ldd (GNU libc) 2.38
Copyright (C) 2023 Free Software Foundation, Inc.
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
Written by Roland McGrath and Ulrich Drepper.""",
                None,
            ],
        ]
    )
    def test_parse_musl_libc(self, ldd_output, expected_musl_libc_version):
        parsed_musl_libc_version = _parse_musl_libc(ldd_output)
        assert expected_musl_libc_version == parsed_musl_libc_version



import json
import os
import textwrap

from conan.api.model import RecipeReference
from conan.internal.paths import CONANFILE
from conan.test.utils.tools import NO_SETTINGS_PACKAGE_ID, TestClient, GenConanfile
from conan.internal.util.files import load


class TestPackageTest:

    def test_basic(self):
        client = TestClient()
        test_package = textwrap.dedent("""
        from conan import ConanFile
        class TestPackage(ConanFile):
            def requirements(self):
                self.requires(self.tested_reference_str)
            def test(self):
                self.output.info("TESTING")
        """)
        client.save({CONANFILE: GenConanfile("hello", "0.1"),
                     "test_package/conanfile.py": test_package})
        client.run("create . --user=lasote --channel=stable")
        assert "hello/0.1@lasote/stable: Created package" in client.out
        client.run("test test_package hello/0.1@lasote/stable")
        assert "hello/0.1@lasote/stable (test package): TESTING" in client.out
        client.run("test hello/0.1@lasote/stable")
        assert "hello/0.1@lasote/stable (test package): TESTING" in client.out

    def test_basic_json(self):
        client = TestClient()
        client.save({CONANFILE: GenConanfile("hello", "0.1"),
                     "test_package/conanfile.py": GenConanfile().with_test("pass")})
        client.run("create . --format=json")
        graph = json.loads(client.stdout)
        assert graph["graph"]["nodes"]["1"]["ref"] == "hello/0.1#a90ba236e5310a473dae9f767a41db91"

    def test_test_only(self):
        test_conanfile = GenConanfile().with_test("pass")
        client = TestClient()
        client.save({CONANFILE: GenConanfile().with_name("hello").with_version("0.1"),
                     "test_package/conanfile.py": test_conanfile})
        client.run("create . --user=lasote --channel=stable")
        client.run("test test_package hello/0.1@lasote/stable")

        assert "Exporting package recipe" not in client.out
        assert "Forced build from source" not in client.out
        assert "Package '%s' created" % NO_SETTINGS_PACKAGE_ID not in client.out
        assert "Forced build from source" not in client.out
        assert "hello/0.1@lasote/stable: Already installed!" in client.out

        client.save({"test_package/conanfile.py": test_conanfile}, clean_first=True)
        client.run("test test_package hello/0.1@lasote/stable")
        assert "hello/0.1@lasote/stable: Configuring sources" not in client.out
        assert "hello/0.1@lasote/stable: Created package" not in client.out
        assert "hello/0.1@lasote/stable: Already installed!" in client.out
        assert "hello/0.1@lasote/stable (test package): Running test()" in client.out

    def test_wrong_version(self):
        test_conanfile = GenConanfile().with_test("pass").with_require("hello/0.2@user/cc")
        client = TestClient()
        client.save({CONANFILE: GenConanfile().with_name("hello").with_version("0.1"),
                     "test_package/conanfile.py": test_conanfile})
        client.run("create . --user=user --channel=channel", assert_error=True)
        assert "Duplicated requirement: hello/0.1@user/channel" in client.out

    def test_other_requirements(self):
        test_conanfile = (GenConanfile().with_require("other/0.2@user2/channel2")
                                        .with_test("pass"))
        client = TestClient()
        other_conanfile = GenConanfile().with_name("other").with_version("0.2")
        client.save({CONANFILE: other_conanfile})
        client.run("export . --user=user2 --channel=channel2")
        client.run("install --requires=other/0.2@user2/channel2 --build='*'")
        client.save({CONANFILE: GenConanfile().with_name("hello").with_version("0.1"),
                     "test_package/conanfile.py": test_conanfile})
        client.run("create . --user=user --channel=channel")
        assert "hello/0.1@user/channel: Created package" in client.out

        # explicit override of user/channel works
        client.run("create . --user=lasote --channel=stable")
        assert "hello/0.1@lasote/stable: Created package" in client.out

    def test_test_with_path_errors(self):
        client = TestClient()
        client.save({"conanfile.txt": "contents"}, clean_first=True)

        # Path with conanfile.txt
        client.run("test conanfile.txt other/0.2@user2/channel2", assert_error=True)

        assert ("A conanfile.py is needed, %s is not acceptable"
                % os.path.join(client.current_folder, "conanfile.txt") in client.out)

        # Path with wrong conanfile path
        client.run("test not_real_dir/conanfile.py other/0.2@user2/channel2", assert_error=True)
        assert ("Conanfile not found at %s"
                % os.path.join(client.current_folder, "not_real_dir", "conanfile.py") in client.out)

    def test_check_version(self):
        client = TestClient()
        client.save({CONANFILE: GenConanfile()})
        client.run("create . --name=dep --version=1.1")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                requires = "dep/1.1"
                def build(self):
                    ref = self.dependencies["dep"].ref
                    self.output.info("BUILD Dep VERSION %s" % ref.version)
            """)
        test_conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                def requirements(self):
                    self.requires(self.tested_reference_str)
                def build(self):
                    ref = self.dependencies["hello"].ref
                    self.output.info("BUILD HELLO VERSION %s" % ref.version)
                def test(self):
                    ref = self.dependencies["hello"].ref
                    self.output.info("TEST HELLO VERSION %s" % ref.version)
            """)
        client.save({"conanfile.py": conanfile,
                     "test_package/conanfile.py": test_conanfile})
        client.run("create . --name=hello --version=0.1")
        assert "hello/0.1: BUILD Dep VERSION 1.1" in client.out
        assert "hello/0.1 (test package): BUILD HELLO VERSION 0.1" in client.out
        assert "hello/0.1 (test package): TEST HELLO VERSION 0.1" in client.out


class TestPackageBuild:
    def test_build_all(self):
        c = TestClient()
        c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
                "dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1"),
                "pkg/test_package/conanfile.py": GenConanfile().with_tool_requires("tool/0.1")
                                                               .with_test("pass")})
        c.run("export tool")
        c.run("export dep")
        c.run("create pkg --build=*")

        c.assert_listed_binary({"dep/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build"),
                                "pkg/0.1": ("59205ba5b14b8f4ebc216a6c51a89553021e82c1", "Build")})
        c.assert_listed_require({"tool/0.1": "Cache"}, build=True, test_package=True)
        c.assert_listed_binary({"tool/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build")},
                               build=True, test_package=True)
        # Note we do NOT rebuild the already built binaries
        c.assert_listed_binary({"dep/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                                "pkg/0.1": ("59205ba5b14b8f4ebc216a6c51a89553021e82c1", "Cache")},
                               test_package=True)

    def test_build_missing(self):
        c = TestClient()
        c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
                "dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1"),
                "pkg/test_package/conanfile.py": GenConanfile().with_tool_requires("tool/0.1")
                                                               .with_test("pass")})
        c.run("export tool")
        c.run("create dep")
        c.run("create pkg --build=missing")
        c.assert_listed_binary({"dep/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                                "pkg/0.1": ("59205ba5b14b8f4ebc216a6c51a89553021e82c1", "Build")})
        c.assert_listed_require({"tool/0.1": "Cache"}, build=True, test_package=True)
        c.assert_listed_binary({"tool/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build")},
                               build=True, test_package=True)
        c.assert_listed_binary({"dep/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                                "pkg/0.1": ("59205ba5b14b8f4ebc216a6c51a89553021e82c1", "Cache")},
                               test_package=True)

    def test_build_test_package_dep(self):
        c = TestClient()
        c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1"),
                "pkg/test_package/conanfile.py": GenConanfile().with_requires("dep/0.1")
                                                               .with_test("pass")})
        c.run("export dep")
        c.run('create pkg --build=missing --build-test=""', assert_error=True)
        c.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build")})
        c.assert_listed_binary({"dep/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Missing"),
                                "pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache")},
                               test_package=True)
        c.run("create pkg --build-test=missing")
        c.assert_listed_binary({"pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build")})
        c.assert_listed_binary({"dep/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build"),
                                "pkg/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache")},
                               test_package=True)


class TestConanTestTest:

    def test_partial_reference(self):
        # Create two packages to test with the same test
        conanfile = '''
from conan import ConanFile

class HelloConan(ConanFile):
    name = "hello"
    version = "0.1"
'''
        client = TestClient()
        client.save({CONANFILE: conanfile})
        client.run("create . --user=conan --channel=stable")
        client.run("create . --user=conan --channel=testing")
        client.run("create . --user=conan --channel=foo")

        def test(conanfile_test, test_reference, path=None):
            path = path or "."
            client.save({os.path.join(path, CONANFILE): conanfile_test}, clean_first=True)
            client.run("test %s %s" % (path, test_reference))

        # Specify a valid name
        test('''
from conan import ConanFile

class HelloTestConan(ConanFile):
    def requirements(self):
        self.requires(self.tested_reference_str)
    def test(self):
        self.output.warning("Tested ok!")
''', "hello/0.1@conan/stable")
        assert "Tested ok!" in client.out

    def test_test_package_env(self):
        client = TestClient()
        conanfile = '''
from conan import ConanFile

class HelloConan(ConanFile):
    name = "hello"
    version = "0.1"
    def package_info(self):
        self.buildenv_info.define("MYVAR", "new/pythonpath/value")

        '''
        test_package = '''
import os, platform
from conan import ConanFile
from conan.tools.env import VirtualBuildEnv

class HelloTestConan(ConanFile):
    generators = "VirtualBuildEnv"

    def requirements(self):
        self.build_requires(self.tested_reference_str)

    def build(self):
        build_env = VirtualBuildEnv(self).vars()
        with build_env.apply():
            assert("new/pythonpath/value" in os.environ["MYVAR"])

    def test(self):
        build_env = VirtualBuildEnv(self).vars()
        with build_env.apply():
            assert("new/pythonpath/value" in os.environ["MYVAR"])
'''

        client.save({"conanfile.py": conanfile, "test_package/conanfile.py": test_package})
        client.run("create . --user=lasote --channel=testing")
        client.run("test test_package hello/0.1@lasote/testing")

    def test_fail_test_package(self):
        client = TestClient()
        conanfile = """
from conan import ConanFile
from conan.tools.files import copy

class HelloConan(ConanFile):
    name = "hello"
    version = "0.1"
    exports_sources = "*"

    def package(self):
        copy(self, "*", self.source_folder, self.package_folder)
"""
        test_conanfile = """
from conan import ConanFile

class HelloReuseConan(ConanFile):
    def requirements(self):
        self.requires(self.tested_reference_str)
    def test(self):
        pass
"""
        client.save({"conanfile.py": conanfile,
                     "FindXXX.cmake": "Hello FindCmake",
                     "test/conanfile.py": test_conanfile})
        client.run("create . --user=lasote --channel=stable")
        ref = RecipeReference.loads("hello/0.1@lasote/stable")
        client.run(f"test test {str(ref)}")
        pref = client.get_latest_package_reference(ref, NO_SETTINGS_PACKAGE_ID)
        assert "Hello FindCmake" == load(os.path.join(client.get_latest_pkg_layout(pref).package(), "FindXXX.cmake"))
        client.save({"FindXXX.cmake": "Bye FindCmake"})
        client.run(f"test test {str(ref)}")  # Test do not rebuild the package
        pref = client.get_latest_package_reference(ref, NO_SETTINGS_PACKAGE_ID)
        assert "Hello FindCmake" == load(os.path.join(client.get_latest_pkg_layout(pref).package(), "FindXXX.cmake"))
        client.run("create . --user=lasote --channel=stable")  # create rebuild the package
        pref = client.get_latest_package_reference(ref, NO_SETTINGS_PACKAGE_ID)
        assert "Bye FindCmake" == load(os.path.join(client.get_latest_pkg_layout(pref).package(), "FindXXX.cmake"))


def test_no_reference_in_test_package():
    client = TestClient()
    test_conanfile = textwrap.dedent("""
        from conan import ConanFile
        import os

        class HelloReuseConan(ConanFile):
            def test(self):
                self.output.warning("At test: {}".format(self.tested_reference_str))
        """)

    client.save({"conanfile.py": GenConanfile(), "test_package/conanfile.py": test_conanfile})
    client.run("create . --name=foo --version=1.0", assert_error=True)
    assert "doesn't declare any requirement, use `self.tested_reference_str` to require the " \
           "package being created" in client.out


def test_tested_reference_str():
    """
    At the test_package/conanfile the variable `self.tested_reference_str` is injected with the
    str of the reference being tested. It is available in all the methods.

    Compatibility with Conan 2.0:
    If the 'test_type' is set to "explicit" the require won't be automatically injected and has to
    be the user the one injecting the require or the build require using the
    `self.tested_reference_str`. This 'test_type' can be removed in 2.0 if we consider it has
    to be always explicit. The recipes will still work in Conan 2.0 because the 'test_type' will be
    ignored.
    """
    client = TestClient()
    test_conanfile = textwrap.dedent("""
    from conan import ConanFile
    import os

    class HelloReuseConan(ConanFile):

        def generate(self):
            self.output.warning("At generate: {}".format(self.tested_reference_str))
            assert len(self.dependencies.values()) == 1
            assert len(self.dependencies.build.values()) == 1

        def build(self):
            self.output.warning("At build: {}".format(self.tested_reference_str))

        def build_requirements(self):
            self.output.warning("At build_requirements: {}".format(self.tested_reference_str))
            self.build_requires(self.tested_reference_str)

        def test(self):
            self.output.warning("At test: {}".format(self.tested_reference_str))
    """)

    client.save({"conanfile.py": GenConanfile(), "test_package/conanfile.py": test_conanfile})
    client.run("create . --name=foo --version=1.0")
    for method in ("generate", "build", "build_requirements", "test"):
        assert "At {}: foo/1.0".format(method) in client.out


def test_folder_output():
    """ the "conan test" command should also follow the test_output layout folder
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("hello", "0.1")})
    c.run("create .")
    c.save({"test_package/conanfile.py": GenConanfile().with_test("pass").with_settings("build_type")
                                                       .with_generator("CMakeDeps")})
    # c.run("create .")
    c.run("test test_package hello/0.1@")
    assert os.path.exists(os.path.join(c.current_folder,
                                       "test_package/hello-config.cmake"))


def test_removing_test_package_build_folder():
    """ The test_package could crash if not cleaning correctly the test_package
    output folder. This will still crassh if the layout is not creating different build folders
    """
    client = TestClient()
    test_package = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Test(ConanFile):
            def requirements(self):
                self.requires(self.tested_reference_str)
            def generate(self):
                assert not os.path.exists("myfile.txt")
                save(self, "myfile.txt", "")
            def layout(self):
                self.folders.build = "mybuild"
                self.folders.generators = "mybuild"
            def test(self):
                pass
            """)
    client.save({"conanfile.py": GenConanfile("pkg", "1.0"),
                 "test_package/conanfile.py": test_package})
    client.run("create .")
    # This was crashing because not cleaned
    client.run("create .")
    assert "Removing previously existing 'test_package' build folder" in client.out


def test_test_package_lockfile_location():
    """ the lockfile should be in the caller cwd
    https://github.com/conan-io/conan/issues/13850
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("dep", "0.1"),
            "test_package/conanfile.py": GenConanfile().with_test("pass")})
    c.run("create . --lockfile-out=myconan.lock")
    assert os.path.exists(os.path.join(c.current_folder, "myconan.lock"))
    c.run("test test_package dep/0.1 --lockfile=myconan.lock --lockfile-out=myconan2.lock")
    assert os.path.exists(os.path.join(c.current_folder, "myconan2.lock"))


def test_package_missing_binary_msg():
    # https://github.com/conan-io/conan/issues/13904
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("dep", "0.1"),
            "test_package/conanfile.py": GenConanfile().with_test("pass")})
    c.run("export .")
    c.run("test test_package dep/0.1", assert_error=True)
    assert "ERROR: Missing binary: dep/0.1" in c.out
    assert "This is a **test_package** missing binary." in c.out
    c.run("test test_package dep/0.1 --build=dep/0.1")
    c.assert_listed_binary({"dep/0.1": (NO_SETTINGS_PACKAGE_ID, "Build")})


def test_test_binary_missing():
    # Trying to reproduce https://github.com/conan-io/conan/issues/14352,
    # without success so far
    c = TestClient()
    c.save({"zlib/conanfile.py": GenConanfile("zlib", "0.1"),
            "openssl/conanfile.py": GenConanfile("openssl", "0.1").with_requires("zlib/0.1"),
            "cmake/conanfile.py": GenConanfile("cmake", "0.1").with_requires("openssl/0.1"),
            "dep/conanfile.py": GenConanfile("dep", "0.1").with_requires("openssl/0.1")
                                                          .with_tool_requires("cmake/0.1"),
            "dep/test_package/conanfile.py": GenConanfile().with_test("pass")})
    c.run("export zlib")
    c.run("export openssl")
    c.run("export cmake")
    c.run("export dep")
    c.run("test dep/test_package dep/0.1 --build=missing")



import os
import textwrap

from conan.test.utils.tools import TestClient


def test_no_copy_source():
    conanfile = textwrap.dedent('''
        from conan import ConanFile
        from conan.tools.files import copy, save, load
        import os

        class ConanFileToolsTest(ConanFile):
            name = "pkg"
            version = "0.1"
            exports_sources = "*"
            no_copy_source = True

            def source(self):
                save(self, "header.h", "artifact contents!")

            def build(self):
                self.output.info("Source files: %s" % load(self,
                                          os.path.join(self.source_folder, "file.h")))
                save(self, "myartifact.lib", "artifact contents!")

            def package(self):
                copy(self, "*", self.source_folder, self.package_folder)
                copy(self, "*", self.build_folder, self.package_folder)
        ''')

    client = TestClient()
    client.save({"conanfile.py": conanfile,
                 "file.h": "myfile.h contents"})
    client.run("create .")
    assert "Source files: myfile.h contents" in client.out
    layout = client.created_layout()
    build_folder = layout.build()
    package_folder = layout.package()

    assert "file.h" not in os.listdir(build_folder)
    assert "header.h" not in os.listdir(build_folder)
    assert "file.h" in os.listdir(package_folder)
    assert "header.h" in os.listdir(package_folder)
    assert "myartifact.lib" in os.listdir(package_folder)
